# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the spec1d directory in idlspec2d.
"""
import glob
import os
import time
from warnings import warn
import numpy as np
from numpy.linalg import solve
try:
    import matplotlib
    matplotlib.use('Agg')
    matplotlib.rcParams['figure.figsize'] = (16.0, 12.0)
    import matplotlib.pyplot as plt
    from matplotlib.font_manager import FontProperties
except ImportError:
    # automodapi can only skip this if it exists.
    FontProperties = None
from astropy import log
from astropy.io import ascii, fits
from . import Pydlspec2dException, Pydlspec2dUserWarning

#
# Used by findspec
#
findspec_cache = None


class HMF(object):
    """Class used to manage data for Heteroscedastic Matrix Factorization (HMF).

    This is a replacement for :func:`~pydl.pydlspec2d.spec1d.pca_solve`.
    It can be called with::

        hmf = HMF(spectra, invvar)
        output = hmf.solve()

    The input spectra should be pre-processed through
    :func:`~pydl.pydlspec2d.spec2d.combine1fiber`.

    Parameters
    ----------
    spectra : array-like
        The input spectral flux, assumed to have a common wavelength and
        redshift system.
    invvar : array-like
        The inverse variance of the spectral flux.
    K : :class:`int`, optional
        The number of dimensions of the factorization (default 4).
    n_iter : :class:`int`, optional
        Number of iterations.
    seed : :class:`int`, optional.
        If set, pass this value to :func:`numpy.random.seed`.
    nonnegative : :class:`bool`, optional
        Set this to ``True`` to perform nonnegative HMF.
    epsilon : :class:`float`, optional
        Regularization parameter.  Set to any non-negative float value to turn
        it on.
    verbose : :class:`bool`, optional
        If ``True``, print extra information.

    Notes
    -----
    See [1]_ and [2]_ for the original derivation of this method.

    The HMF iteration is initialized using :func:`~scipy.cluster.vq.kmeans`,
    which itself uses random numbers to initialize its state.  If you need
    to ensure reproducibility, call :func:`numpy.random.seed` before
    initializing HMF.

    The current algorithm cannot handle input data that contain *columns*
    of zeros.  Columns of this type need to be *carefully* removed from the
    input data.  This could also result in the output data having a different
    size compared to the input data.

    References
    ----------
    .. [1] `Tsalmantza, P., Decarli, R., Dotti, M., Hogg, D. W., 2011 ApJ 738, 20
        <https://ui.adsabs.harvard.edu/abs/2011ApJ...738...20T/abstract>`_
    .. [2] `Tsalmantza, P., Hogg, D. W., 2012 ApJ 753, 122
        <https://ui.adsabs.harvard.edu/abs/2012ApJ...753..122T/abstract>`_
    """

    def __init__(self, spectra, invvar, K=4, n_iter=None, seed=None,
                 nonnegative=False, epsilon=None, verbose=False):
        self.spectra = spectra
        self.invvar = invvar
        self.K = K
        if n_iter is None:
            if nonnegative:
                self.n_iter = 2048
            else:
                self.n_iter = 20
        else:
            self.n_iter = int(n_iter)
        self.seed = seed
        self.nonnegative = nonnegative
        self.epsilon = epsilon
        self.verbose = verbose
        self.a = None
        self.g = None
        if self.verbose:
            log.setLevel('DEBUG')
        return

    def solve(self):
        """Process the inputs.

        Returns
        -------
        :class:`dict`
            The HMF solution.
        """
        if len(self.spectra.shape) == 1:
            nobj = 1
            npix = self.spectra.shape[0]
        else:
            nobj, npix = self.spectra.shape
        log.info("Building HMF from %d object spectra.", nobj)
        fluxdict = dict()
        #
        # If there is only one object spectrum, then all we can do is return it.
        #
        if nobj == 1:
            fluxdict['flux'] = self.spectra.astype('f')
            return fluxdict
        a, g = self.iterate()
        fluxdict['acoeff'] = a
        fluxdict['flux'] = g
        return fluxdict

    def model(self):
        """Compute the model.
        """
        return np.dot(self.a, self.g)

    def resid(self):
        """Compute residuals.
        """
        return self.spectra - self.model()

    def chi(self):
        r"""Compute :math:`\chi`, the scaled residual.
        """
        return self.resid() * np.sqrt(self.invvar)

    def penalty(self):
        """Compute penalty for non-smoothness.
        """
        if self.epsilon is None:
            return 0.0
        return self.epsilon * np.sum(np.diff(self.g)**2)

    def badness(self):
        r"""Compute :math:`\chi^2`, including possible non-smoothness penalty.
        """
        return np.sum(self.chi()**2) + self.penalty()

    def normbase(self):
        """Apply standard component normalization.
        """
        return np.sqrt((self.g**2).mean(1))

    def astep(self):
        """Update for coefficients at fixed component spectra.
        """
        N, M = self.spectra.shape
        K, M = self.g.shape
        a = np.zeros((N, K), dtype=self.g.dtype)
        for i in range(N):
            Gi = np.zeros((K, K), dtype=self.g.dtype)
            for k in range(K):
                for kp in range(k, K):
                    Gi[k, kp] = np.sum(self.g[k, :] * self.g[kp, :] *
                                       self.invvar[i, :])
                    if kp > k:
                        Gi[kp, k] = Gi[k, kp]
            Fi = np.dot(self.g, self.spectra[i, :]*self.invvar[i, :])
            a[i, :] = solve(Gi, Fi)
        return a

    def gstep(self):
        """Update for component spectra at fixed coefficients.
        """
        N, M = self.spectra.shape
        N, K = self.a.shape
        g = np.zeros((K, M), dtype=self.a.dtype)
        e = np.zeros(self.g.shape, dtype=self.g.dtype)
        d = np.zeros((K, K, M), dtype=self.a.dtype)
        if self.epsilon is not None and self.epsilon > 0:
            foo = self.epsilon * np.eye(K, dtype=self.a.dtype)
            for l in range(M):
                d[:, :, l] = foo
                if l > 0 and l < M-1:
                    d[:, :, l] *= 2
            # d[:, :, 0] = foo
            # d[:, :, 1:M-1] = 2*foo
            # d[:, :, M-1] = foo
            e[:, 0] = self.epsilon*self.g[:, 1]
            e[:, 1:M-1] = self.epsilon*(self.g[:, 0:M-2] + self.g[:, 2:M])
            e[:, M-1] = self.epsilon*self.g[:, M-2]
        for j in range(M):
            Aj = np.zeros((K, K), dtype=self.a.dtype)
            for k in range(K):
                for kp in range(k, K):
                    Aj[k, kp] = np.sum(self.a[:, k] * self.a[:, kp] *
                                       self.invvar[:, j])
                    if kp > k:
                        Aj[kp, k] = Aj[k, kp]
            Aj += d[:, :, j]
            Fj = (np.dot(self.a.T, self.spectra[:, j]*self.invvar[:, j]) +
                  e[:, j])
            g[:, j] = solve(Aj, Fj)
        return g

    def astepnn(self):
        """Non-negative update for coefficients at fixed component spectra.
        """
        numerator = np.dot(self.spectra*self.invvar, self.g.T)
        denominator = np.dot(np.dot(self.a, self.g)*self.invvar, self.g.T)
        return self.a*(numerator/denominator)

    def gstepnn(self):
        """Non-negative update for component spectra at fixed coefficients.
        """
        K, M = self.g.shape
        numerator = np.dot(self.a.T, (self.spectra*self.invvar))
        if self.epsilon is not None and self.epsilon > 0:
            e = np.zeros(self.g.shape, dtype=self.g.dtype)
            e[:, 0] = self.epsilon*self.g[:, 1]
            e[:, 1:M-1] = self.epsilon*(self.g[:, 0:M-2] + self.g[:, 2:M])
            e[:, M-1] = self.epsilon*self.g[:, M-2]
            numerator += e
        denominator = np.dot(self.a.T, np.dot(self.a, self.g)*self.invvar)
        if self.epsilon is not None and self.epsilon > 0:
            d = self.epsilon*self.g.copy()
            d[:, 1:M-1] *= 2
            denominator += d
        return self.g*(numerator/denominator)

    def reorder(self):
        """Reorder and rotate basis analogous to PCA.
        """
        from numpy.linalg import eigh
        l, U = eigh(np.dot(self.a.T, self.a))
        return (np.dot(self.a, U), np.dot(U.T, self.g))

    def iterate(self):
        """Handle the HMF iteration.

        Returns
        -------
        :class:`tuple` of :class:`numpy.ndarray`
            The fitting coefficients and fitted functions, respectively.
        """
        from scipy.cluster.vq import kmeans, whiten
        from ..pydlutils.math import find_contiguous
        N, M = self.spectra.shape
        #
        # Make spectra non-negative
        #
        if self.nonnegative:
            self.spectra[self.spectra < 0] = 0
            self.invvar[self.spectra < 0] = 0
        #
        # Detect and fix very bad columns.
        #
        si = self.spectra * self.invvar
        if (self.spectra.sum(0) == 0).any():
            log.warn("Columns of zeros detected in spectra!")
        if (self.invvar.sum(0) == 0).any():
            log.warn("Columns of zeros detected in invvar!")
        if (si.sum(0) == 0).any():
            log.warn("Columns of zeros detected in spectra*invvar!")
        zerocol = ((self.spectra.sum(0) == 0) | (self.invvar.sum(0) == 0) |
                   (si.sum(0) == 0))
        n_zero = zerocol.sum()
        if n_zero > 0:
            log.warn("Found %d bad columns in input data!", n_zero)
        #
        # Find the largest set of contiguous pixels
        #
        goodcol = find_contiguous(~zerocol)
        self.spectra = self.spectra[:, goodcol]
        self.invvar = self.invvar[:, goodcol]
        # si = si[:, goodcol]
        # newloglam = fullloglam[goodcol]
        #
        # Initialize g matrix with kmeans
        #
        if self.seed is not None:
            np.random.seed(self.seed)
        whitespectra = whiten(self.spectra)
        log.debug(whitespectra[0:3, 0:3])
        self.g, foo = kmeans(whitespectra, self.K)
        self.g /= np.repeat(self.normbase(), M).reshape(self.g.shape)
        log.debug(self.g[0:3, 0:3])
        #
        # Initialize a matrix
        #
        self.a = np.outer(np.sqrt((self.spectra**2).mean(1)),
                          np.repeat(1.0/self.K, self.K))
        if self.nonnegative:
            for k in range(128):
                self.a = self.astepnn()
        #
        # Iterate!
        #
        t0 = time.time()
        for m in range(self.n_iter):
            log.info("Starting iteration #%4d.", m+1)
            if self.nonnegative:
                self.a = self.astepnn()
                self.g = self.gstepnn()
            else:
                self.a = self.astep()
                self.g = self.gstep()
                self.a, self.g = self.reorder()
            norm = self.normbase()
            self.g /= np.repeat(norm, M).reshape(self.g.shape)
            self.a = (self.a.T*np.repeat(norm, N).reshape(self.K, N)).T
            log.debug(self.a[0:3, 0:3])
            log.debug(self.g[0:3, 0:3])
            log.debug("Chi**2 after iteration #%4d = %f.", m+1, self.badness())
            log.info("The elapsed time for iteration #%4d is %6.2f s.", m+1, time.time()-t0)
        return (self.a, self.g)


def findspec(*args, **kwargs):
    """Find SDSS/BOSS spectra that match a given RA, Dec.

    Parameters
    ----------
    ra, dec : array-like, optional
        If set, the first two positional arguments will be interpreted as
        RA, Dec.
    best : :class:`bool`, optional
        If set, return only the best match for each input RA, Dec.
    infile : :class:`str`, optional
        If set, read RA, Dec data from this file.
    outfile : :class:`str`, optional
        If set, print match data to this file.
    print : :class:`bool`, optional
        If set, print the match data to the console.
    run1d : :class:`str`, optional
        Override the value of :envvar:`RUN1D`.
    run2d : :class:`str`, optional
        Override the value of :envvar:`RUN2D`.
    sdss : :class:`bool`, optional
        If set, search for SDSS-I/II spectra instead of BOSS spectra.
    searchrad : :class:`float`, optional
        Search for spectra in this radius around given RA, Dec.
        Default is 3 arcsec.
    topdir : :class:`str`, optional
        If set, override the value of :envvar:`SPECTRO_REDUX`
        or :envvar:`BOSS_SPECTRO_REDUX`.

    Returns
    -------
    :class:`dict`
        A dictionary containing plate, MJD, fiber, etc.
    """
    from .. import uniq
    from ..pydlutils.misc import struct_print
    from ..pydlutils.spheregroup import spherematch
    global findspec_cache
    #
    # Set up default values
    #
    if 'sdss' in kwargs:
        if 'topdir' in kwargs:
            topdir = kwargs['topdir']
        else:
            topdir = os.environ['SPECTRO_REDUX']
        if 'run2d' in kwargs:
            run2d = str(kwargs['run2d'])
        else:
            run2d = '26'
        run1d = ''
    else:
        if 'topdir' in kwargs:
            topdir = kwargs['topdir']
        else:
            topdir = os.environ['BOSS_SPECTRO_REDUX']
        if 'run2d' in kwargs:
            run2d = str(kwargs['run2d'])
        else:
            run2d = os.environ['RUN2D']
        if 'run1d' in kwargs:
            run1d = str(kwargs['run1d'])
        else:
            run1d = os.environ['RUN1D']
    if findspec_cache is None:
        findspec_cache = {'lasttopdir': topdir, 'plist': None}
    if (findspec_cache['plist'] is None or topdir != findspec_cache['lasttopdir']):
        findspec_cache['lasttopdir'] = topdir
        platelist_file = os.path.join(topdir, "platelist.fits")
        plates_files = glob.glob(os.path.join(topdir, "plates-*.fits"))
        plist = None
        if os.path.exists(platelist_file):
            platelist = fits.open(platelist_file)
            plist = platelist[1].data
            platelist.close()
        if len(plates_files) > 0:
            plates = fits.open(plates_files[0])
            plist = plates[1].data
            plates.close()
        if plist is None:
            raise Pydlspec2dException("Plate list (platelist.fits or plates-*.fits) not found in {0}.".format(topdir))
        else:
            findspec_cache['plist'] = plist
    qdone = plist.field('STATUS1D') == 'Done'
    qdone2d = plist.field('RUN2D').strip() == run2d
    if run1d == '':
        qdone1d = np.ones(plist.size, dtype='bool')
    else:
        qdone1d = plist.field('RUN1D').strip() == run1d
    qfinal = qdone & qdone2d & qdone1d
    if not qfinal.any():
        warn("No reduced plates!", Pydlspec2dUserWarning)
        return None
    idone = np.arange(plist.size)[qfinal]
    #
    # If there are positional arguments, interpret these as RA, Dec
    #
    if len(args) == 2:
        ra = args[0]
        dec = args[1]
    #
    # Read RA, Dec from infile if set
    #
    if 'infile' in kwargs:
        infile_data = ascii.read(kwargs['infile'], names=['ra', 'dec'])
        ra = infile_data["ra"].data
        dec = infile_data["dec"].data
    if 'searchrad' in kwargs:
        searchrad = float(kwargs['searchrad'])
    else:
        searchrad = 3.0/3600.0
    #
    # Create output structure
    #
    slist_type = np.dtype([('PLATE', 'i4'), ('MJD', 'i4'), ('FIBERID', 'i4'),
                          ('RA', 'f8'), ('DEC', 'f8'), ('MATCHRAD', 'f8')])
    #
    # Match all plates with objects
    #
    imatch1, itmp, dist12 = spherematch(ra, dec, plist[qfinal].field('RACEN'),
                                        plist[qfinal].field('DECCEN'),
                                        searchrad+1.55, maxmatch=0)
    if imatch1.size == 0:
        warn("No matching plates found.", Pydlspec2dUserWarning)
        return None
    imatch2 = idone[itmp]
    #
    # Read all relevant plates
    #
    try:
        n_total = plist.field('N_TOTAL')
    except KeyError:
        n_total = np.zeros(plist.size, dtype='i4') + 640
    iplate = imatch2[uniq(imatch2, imatch2.argsort())]
    i0 = 0
    plugmap = np.zeros(n_total[iplate].sum(), dtype=[('PLATE', 'i4'),
                       ('MJD', 'i4'), ('FIBERID', 'i4'),
                       ('RA', 'd'), ('DEC', 'd')])
    for i in range(iplate.size):
        spplate = readspec(plist[iplate[i]].field('PLATE'),
                           mjd=plist[iplate[i]].field('MJD'),
                           topdir=topdir,
                           run2d=run2d, run1d=run1d)
        index_to = i0 + np.arange(n_total[iplate[i]], dtype='i4')
        plugmap['PLATE'][index_to] = plist[iplate[i]].field('PLATE')
        plugmap['MJD'][index_to] = plist[iplate[i]].field('MJD')
        plugmap['FIBERID'][index_to] = spplate['plugmap']['FIBERID']
        plugmap['RA'][index_to] = spplate['plugmap']['RA']
        plugmap['DEC'][index_to] = spplate['plugmap']['DEC']
        i0 += n_total[iplate[i]]
    i1, i2, d12 = spherematch(ra, dec, plugmap['RA'], plugmap['DEC'],
                              searchrad, maxmatch=0)
    if i1.size == 0:
        warn('No matching objects found.', Pydlspec2dUserWarning)
        return None
    if 'best' in kwargs:
        #
        # Return only best match per object
        #
        slist = np.zeros(ra.size, dtype=slist_type)
        spplate = readspec(plugmap[i2]['PLATE'],
                           plugmap[i2]['FIBERID'],
                           mjd=plugmap[i2]['MJD'],
                           topdir=topdir,
                           run2d=run2d, run1d=run1d)
        sn = spplate['zans']['SN_MEDIAN']
        isort = (i1 + np.where(sn > 0, sn, 0)/(sn+1.0).max()).argsort()
        i1 = i1[isort]
        i2 = i2[isort]
        d12 = d12[isort]
        iuniq = uniq(i1)
        slist[i1[iuniq]]['PLATE'] = plugmap[i2[iuniq]]['PLATE']
        slist[i1[iuniq]]['MJD'] = plugmap[i2[iuniq]]['MJD']
        slist[i1[iuniq]]['FIBERID'] = plugmap[i2[iuniq]]['FIBERID']
        slist[i1[iuniq]]['RA'] = plugmap[i2[iuniq]]['RA']
        slist[i1[iuniq]]['DEC'] = plugmap[i2[iuniq]]['DEC']
        slist[i1[iuniq]]['MATCHRAD'] = d12[iuniq]
    else:
        #
        # Return all matches
        #
        slist = np.zeros(i1.size, dtype=slist_type)
        slist['PLATE'] = plugmap[i2]['PLATE']
        slist['MJD'] = plugmap[i2]['MJD']
        slist['FIBERID'] = plugmap[i2]['FIBERID']
        slist['RA'] = plugmap[i2]['RA']
        slist['DEC'] = plugmap[i2]['DEC']
        slist['MATCHRAD'] = d12
    #
    # Print to terminal or output file
    #
    if 'print' in kwargs:
        _ = struct_print(slist)
    if 'outfile' in kwargs:
        _ = struct_print(slist, filename=kwargs['outfile'])
    return slist


def latest_mjd(plate, **kwargs):
    """Find the most recent MJD associated with a plate.

    Parameters
    ----------
    plate : :class:`int` or :class:`numpy.ndarray`
        The plate(s) to examine.

    Returns
    -------
    :class:`numpy.ndarray`
        An array of MJD values for each plate.
    """
    import re
    if isinstance(plate, (int,)) or plate.shape == ():
        platevec = np.array([plate], dtype='i4')
    else:
        platevec = plate
    mjd = np.zeros(len(platevec), dtype='i4')
    mjdre = re.compile(r'spPlate-[0-9]{4}-([0-9]{5}).fits')
    unique_plates = np.unique(platevec)
    paths = spec_path(unique_plates, **kwargs)
    for p, q in zip(paths, unique_plates):
        plateglob = "{0}/spPlate-{1:04d}-*.fits".format(p, q)
        bigmjd = 0
        for f in glob.glob(plateglob):
            thismjd = int(mjdre.search(f).groups()[0])
            if thismjd > bigmjd:
                bigmjd = thismjd
        mjd[platevec == q] = bigmjd
    return mjd


def number_of_fibers(plate, **kwargs):
    """Returns the total number of fibers per plate.

    Parameters
    ----------
    plate : :class:`int` or :class:`numpy.ndarray`
        The plate(s) to examine.

    Returns
    -------
    :class:`numpy.ndarray`
        The number of fibers on each plate.
    """
    #
    # Get mjd values
    #
    if isinstance(plate, (int,)) or plate.shape == ():
        platevec = np.array([plate], dtype='i4')
    else:
        platevec = plate
    mjd = latest_mjd(plate, **kwargs)
    nfiber = np.zeros(mjd.size, dtype='i4')
    #
    # SDSS-I,II plates
    #
    nfiber[mjd < 55025] = 640
    #
    # Short circuit if we're done.
    #
    if (nfiber == 640).all():
        return nfiber
    #
    # Not all BOSS plates have 1000 fibers
    #
    if 'path' in kwargs:
        platelistpath = os.path.join(kwargs['path'], 'platelist.fits')
    else:
        platelistpath = os.path.join(os.environ['BOSS_SPECTRO_REDUX'], 'platelist.fits')
    platelist = fits.open(platelistpath)
    platentotal = platelist[1].data.field('N_TOTAL')
    plateplate = platelist[1].data.field('PLATE')
    platemjd = platelist[1].data.field('MJD')
    platerun2d = platelist[1].data.field('RUN2D')
    platerun1d = platelist[1].data.field('RUN1D')
    platelist.close()
    if 'run2d' in kwargs:
        run2d = kwargs['run2d']
    else:
        run2d = os.environ['RUN2D']
    if 'run1d' in kwargs:
        run1d = kwargs['run1d']
    else:
        run1d = os.environ['RUN1D']
    for k in range(mjd.size):
        nfiber[k] = platentotal[(plateplate == platevec[k]) &
                                (platemjd == mjd[k]) &
                                (platerun2d == run2d) &
                                (platerun1d == run1d)]
    return nfiber


def pca_solve(newflux, newivar, maxiter=0, niter=10, nkeep=3,
              nreturn=None, verbose=False):
    """Replacement for idlspec2d pca_solve.pro.

    Parameters
    ----------
    newflux : array-like
        The input spectral flux, assumed to have a common wavelength and
        redshift system.
    newivar : array-like
        The inverse variance of the spectral flux.
    maxiter : :class:`int`, optional
        Stop PCA+reject iterations after this number.
    niter : :class:`int`, optional
        Stop PCA iterations after this number.
    nkeep : :class:`int`, optional
        Number of PCA components to keep.
    nreturn : :class:`int`, optional
        Number of PCA components to return, usually the same as `nkeep`.
    verbose : :class:`bool`, optional
        If ``True``, print extra information.

    Returns
    -------
    :class:`dict`
        The PCA solution.
    """
    from .. import pcomp
    from ..pydlutils.math import computechi2, djs_reject
    if verbose:
        log.setLevel('DEBUG')
    if nreturn is None:
        nreturn = nkeep
    if len(newflux.shape) == 1:
        nobj = 1
        npix = newflux.shape[0]
    else:
        nobj, npix = newflux.shape
    log.info("Building PCA from %d object spectra.", nobj)
    nzi = newivar.nonzero()
    first_nonzero = (np.arange(nobj, dtype=nzi[0].dtype),
                     np.array([nzi[1][nzi[0] == k].min() for k in range(nobj)]))
    #
    # Construct the synthetic weight vector, to be used when replacing the
    # low-S/N object pixels with the reconstructions.
    #
    synwvec = np.ones((npix,), dtype='d')
    for ipix in range(npix):
        indx = newivar[:, ipix] != 0
        if indx.any():
            synwvec[ipix] = newivar[indx, ipix].mean()
    fluxdict = dict()
    #
    # If there is only one object spectrum, then all we can do is return it.
    #
    if nobj == 1:
        fluxdict['flux'] = newflux.astype('f')
        return fluxdict
    #
    # Rejection iteration loop.
    #
    qdone = 0
    iiter = 0
    #
    # Begin with all points good.
    #
    outmask = None
    inmask = newivar != 0
    ymodel = None
    # emevecs, emevals = pydlutils.empca(newflux, inmask)
    # fluxdict['emevecs'] = emevecs
    # fluxdict['emevals'] = emeveals
    while qdone == 0 and iiter <= maxiter:
        log.debug('starting djs_reject')
        outmask, qdone = djs_reject(newflux, ymodel, inmask=inmask,
                                    outmask=outmask, invvar=newivar)
        log.debug('finished with djs_reject')
        #
        # Iteratively do the PCA solution
        #
        filtflux = newflux.copy()
        acoeff = np.zeros((nobj, nkeep), dtype='d')
        t0 = time.time()
        for ipiter in range(niter):
            #
            # We want to get these values from the pcomp routine.
            #
            # eigenval = 1
            # coeff = 1
            flux0 = np.tile(filtflux[first_nonzero], npix).reshape(npix, nobj).transpose()
            # flux0 = np.tile(filtflux, npix).reshape(npix, nobj).transpose()
            totflux = np.absolute(filtflux - flux0).sum(1)
            goodobj = totflux > 0
            if goodobj.all():
                tmp = pcomp(filtflux.T)  # , standardize=True)
                pres = tmp.derived
                eigenval = tmp.eigenvalues
            else:
                tmp = pcomp(filtflux[goodobj, :].T)  # , standardize=True)
                pres = np.zeros((nobj, npix), dtype='d')
                pres[goodobj, :] = tmp.derived
                eigenval = np.zeros((nobj,), dtype='d')
                eigenval[goodobj] = tmp.eigenvalues
            maskivar = newivar * outmask
            sqivar = np.sqrt(maskivar)
            for iobj in range(nobj):
                out = computechi2(newflux[iobj, :], sqivar[iobj, :],
                                  pres[:, 0:nkeep])
                filtflux[iobj, :] = (maskivar[iobj, :] * newflux[iobj, :] +
                                     synwvec*out.yfit) / (maskivar[iobj, :] +
                                                          synwvec)
                acoeff[iobj, :] = out.acoeff
            log.info("The elapsed time for iteration #%2d is %6.2f s.", ipiter+1, time.time()-t0)
        #
        # Now set ymodel for rejecting points.
        #
        ymodel = np.dot(acoeff, pres[:, 0:nkeep].T)
        iiter += 1
    if nobj == 1:
        usemask = outmask
    else:
        usemask = outmask.sum(0)
    fluxdict['usemask'] = usemask
    fluxdict['outmask'] = outmask
    fluxdict['flux'] = pres[:, 0:nreturn].transpose().astype('f')
    fluxdict['eigenval'] = eigenval[0:nreturn]
    fluxdict['acoeff'] = acoeff
    return fluxdict


def plot_eig(filename, title='Unknown'):
    """Plot spectra from an eigenspectra/template file.

    Parameters
    ----------
    filename : :class:`str`
        Name of a FITS file containing eigenspectra/templates.
    title : :class:`str`, optional
        Title to put on the plot.

    Raises
    ------
    :exc:`ValueError`
        If an unknown template type was input in `filename`.
    """
    #
    # Set title based on filename
    #
    if title == 'Unknown':
        if filename.find('Gal') > 0:
            title = 'Galaxies: Eigenspectra'
        elif filename.find('QSO') > 0:
            title = 'QSOs: Eigenspectra'
        elif filename.find('Star') > 0:
            title = 'Stars: Eigenspectra'
        elif filename.find('CVstar') > 0:
            title = 'CV Stars: Eigenspectra'
        else:
            raise ValueError('Unknown template type!')
    base, ext = filename.split('.')
    spectrum = fits.open(filename)
    newloglam0 = spectrum[0].header['COEFF0']
    objdloglam = spectrum[0].header['COEFF1']
    spectro_data = spectrum[0].data
    spectrum.close()
    (neig, ndata) = spectro_data.shape
    newloglam = np.arange(ndata) * objdloglam + newloglam0
    lam = 10.0**newloglam
    fig = plt.figure(dpi=100)
    ax = fig.add_subplot(111)
    colorvec = ['k', 'r', 'g', 'b', 'm', 'c']
    for l in range(neig):
        _ = ax.plot(lam, spectro_data[l, :],
                    colorvec[l % len(colorvec)]+'-', linewidth=1)
    ax.set_xlabel(r'Wavelength [$\AA$]')
    ax.set_ylabel('Flux [Arbitrary Units]')
    ax.set_title(title)
    # ax.set_xlim([3500.0,10000.0])
    # ax.set_ylim([-400.0,500.0])
    # fig.savefig(base+'.zoom.png')
    fig.savefig(base+'.png')
    plt.close(fig)
    return


def readspec(platein, mjd=None, fiber=None, **kwargs):
    """Read SDSS/BOSS spec2d & spec1d files.

    Parameters
    ----------
    platein : :class:`int` or :class:`numpy.ndarray`
        Plate number(s).
    mjd : :class:`int` or :class:`numpy.ndarray`, optional
        MJD numbers.  If not provided, they will be calculated by
        :func:`latest_mjd`.
    fiber : array-like, optional
        Fibers to read.  If not set, all fibers from all plates will be
        returned.
    topdir : :class:`str`, optional
        Override the value of :envvar:`BOSS_SPECTRO_REDUX`.
    run2d : :class:`str`, optional
        Override the value of :envvar:`RUN2D`.
    run1d : :class:`str`, optional
        Override the value of :envvar:`RUN1D`.
    path : :class:`str`, optional
        Override all path information with this directory name.
    align : :class:`bool`, optional
        If set, align all the spectra in wavelength.
    znum : :class:`int`, optional
        If set, return the znum-th best fit reshift fit, instead of the best.

    Returns
    -------
    :class:`dict`
        A dictionary containing the data read.
    """
    try:
        nplate = len(platein)
        plate = platein
    except TypeError:
        nplate = 1
        plate = np.array([platein], dtype='i4')
    if 'run2d' in kwargs:
        run2d = kwargs['run2d']
    else:
        run2d = os.environ['RUN2D']
    if 'run1d' in kwargs:
        run1d = kwargs['run1d']
    else:
        run1d = os.environ['RUN1D']
    if fiber is None:
        #
        # Read all fibers
        #
        nfibers = number_of_fibers(plate, **kwargs)
        total_fibers = nfibers.sum()
        platevec = np.zeros(total_fibers, dtype='i4')
        fibervec = np.zeros(total_fibers, dtype='i4')
        k = 0
        for p in np.unique(plate):
            n = np.unique(nfibers[plate == p])[0]
            platevec[k:k+n] = p
            fibervec[k:k+n] = np.arange(n) + 1
            k += n
    else:
        try:
            nfiber = len(fiber)
        except TypeError:
            nfiber = 1
        if nplate > 1 and nfiber > 1 and nplate != nfiber:
            raise TypeError("Plate & Fiber must have the same length!")
        if nplate > 1:
            platevec = np.array(plate, dtype='i4')
        else:
            platevec = np.zeros(nfiber, dtype='i4') + plate
        if nfiber > 1:
            fibervec = np.array(fiber, dtype='i4')
        else:
            fibervec = np.zeros(nplate, dtype='i4') + fiber
    if mjd is None:
        mjdvec = latest_mjd(platevec, **kwargs)
    else:
        try:
            nmjd = len(mjd)
        except TypeError:
            nmjd = 1
        if nmjd != nplate:
            raise TypeError("Plate & MJD must have the same length!")
        mjdvec = np.zeros(nplate, dtype='i4') + mjd
    #
    # Now select unique plate-mjd combinations & read them
    #
    pmjd = ((np.array(platevec, dtype='u8') << 16) +
            np.array(mjdvec, dtype='u8'))
    # log.debug(pmjd)
    upmjd = np.unique(pmjd)
    zupmjd = list(zip(upmjd >> 16, upmjd & ((1 << 16) - 1)))
    # log.debug(zupmjd)
    spplate_data = dict()
    hdunames = ('flux', 'invvar', 'andmask', 'ormask', 'disp', 'plugmap',
                'sky', 'loglam',)
    for thisplate, thismjd in zupmjd:
        # thisplate = int(p>>16)
        # thismjd = int(np.bitwise_and(p, (1<<16)-1))
        pmjdindex = ((platevec == thisplate) &
                     (mjdvec == thismjd)).nonzero()[0]
        thisfiber = fibervec[pmjdindex]
        # log.debug(type(thisplate), type(thismjd))
        # log.debug(repr(thisfiber))
        # log.debug(type(thisfiber))
        pmjdstr = "{0:04d}-{1:05d}".format(int(thisplate), int(thismjd))
        if 'path' in kwargs:
            sppath = [kwargs['path']]
        else:
            sppath = spec_path(thisplate, run2d=run2d)
        spfile = os.path.join(sppath[0], "spPlate-{0}.fits".format(pmjdstr))
        log.info(spfile)
        spplate = fits.open(spfile)
        #
        # Get wavelength coefficients from primary header
        #
        npix = spplate[0].header['NAXIS1']
        c0 = spplate[0].header['COEFF0']
        c1 = spplate[0].header['COEFF1']
        coeff0 = np.zeros(thisfiber.size, dtype='d') + c0
        coeff1 = np.zeros(thisfiber.size, dtype='d') + c1
        loglam0 = c0 + c1*np.arange(npix, dtype='d')
        loglam = np.resize(loglam0, (thisfiber.size, npix))
        #
        # Read the data images
        #
        for k in range(len(hdunames)):
            if hdunames[k] == 'loglam':
                tmp = loglam
            else:
                try:
                    tmp = spplate[k].data[thisfiber-1, :]
                except IndexError:
                    tmp = spplate[k].data[thisfiber-1]
            if hdunames[k] not in spplate_data:
                if k == 0:
                    allpmjdindex = pmjdindex
                    allcoeff0 = coeff0
                    allcoeff1 = coeff1
                #
                # Put the data into the return structure
                #
                if hdunames[k] == 'plugmap':
                    spplate_data['plugmap'] = dict()
                    for c in spplate[k].columns.names:
                        spplate_data['plugmap'][c] = tmp[c]
                else:
                    spplate_data[hdunames[k]] = tmp
            else:
                #
                # Append data
                #
                if k == 0:
                    allpmjdindex = np.concatenate((allpmjdindex, pmjdindex))
                    if 'align' in kwargs:
                        mincoeff0 = min(allcoeff0)
                        if mincoeff0 == 0 and coeff0[0] > 0:
                            allcoeff0 = coeff0[0]
                            allcoeff1 = coeff1[1]
                        if mincoeff0 > 0 and coeff0[0] == 0:
                            coeff0 = mincoeff0
                            coeff1 = allcoeff1[0]
                        ps = np.floor((coeff0[0] - mincoeff0)/coeff1[0] + 0.5)
                        if ps > 0:
                            coeff0 = coeff0 - ps*coeff1
                        else:
                            allcoeff0 = allcoeff0 + ps*allcoeff1
                    else:
                        ps = 0
                    allcoeff0 = np.concatenate((allcoeff0, coeff0))
                    allcoeff1 = np.concatenate((allcoeff1, coeff1))
                if hdunames[k] == 'plugmap':
                    for c in spplate[5].columns.names:
                        spplate_data['plugmap'][c] = np.concatenate(
                            (spplate_data['plugmap'][c], tmp[c]))
                else:
                    spplate_data[hdunames[k]] = spec_append(spplate_data[hdunames[k]], tmp, pixshift=ps)
        spplate.close()
        #
        # Read photoPlate information, if available
        #
        photofile = os.path.join(sppath[0],
                                 "photoPlate-{0}.fits".format(pmjdstr))
        if not os.path.exists(photofile):
            #
            # Hmm, maybe this is an SDSS-I,II plate
            #
            photofile = os.path.join(os.environ['SPECTRO_MATCH'], run2d,
                                     os.path.basename(os.environ['PHOTO_RESOLVE']),
                                     "{0:04d}".format(int(thisplate)),
                                     "photoPlate-{0}.fits".format(pmjdstr))
        if os.path.exists(photofile):
            photop = fits.open(photofile)
            tmp = photop[1].data[thisfiber-1]
            if 'tsobj' not in spplate_data:
                spplate_data['tsobj'] = dict()
                for c in photop[1].columns.names:
                    spplate_data['tsobj'][c] = tmp[c]
            else:
                for c in photop[1].columns.names:
                    spplate_data['tsobj'][c] = np.concatenate(
                        (spplate_data['tsobj'][c], tmp[c]))
            photop.close()

        #
        # Read redshift information, if available.
        #
        if 'znum' in kwargs:
            zfile = os.path.join(sppath[0], run1d,
                                 "spZall-{0}.fits".format(pmjdstr))
        else:
            zfile = os.path.join(sppath[0], run1d,
                                 "spZbest-{0}.fits".format(pmjdstr))
        if os.path.exists(zfile):
            spz = fits.open(zfile)
            if 'znum' in kwargs:
                nper = spz[0].header['DIMS0']
                zfiber = (thisfiber-1)*nper + kwargs['znum'] - 1
            else:
                zfiber = thisfiber
            tmp = spz[1].data[zfiber-1]
            if 'zans' not in spplate_data:
                spplate_data['zans'] = dict()
                for c in spz[1].columns.names:
                    spplate_data['zans'][c] = tmp[c]
            else:
                for c in spz[1].columns.names:
                    spplate_data['zans'][c] = np.concatenate(
                        (spplate_data['zans'][c], tmp[c]))
            spz.close()
    #
    # Reorder the data.  At this point allpmjdindex is an index for which
    # fiber[allpmjdindex] == spplate['plugmap']['FIBERID'], so we have to
    # reverse this mapping.
    #
    j = allpmjdindex.argsort()
    for k in spplate_data:
        if isinstance(spplate_data[k], dict):
            for c in spplate_data[k]:
                if spplate_data[k][c].ndim == 2:
                    spplate_data[k][c] = spplate_data[k][c][j, :]
                else:
                    spplate_data[k][c] = spplate_data[k][c][j]
        else:
            spplate_data[k] = spplate_data[k][j, :]
    allcoeff0 = allcoeff0[j]
    allcoeff1 = allcoeff1[j]
    #
    # If necessary, recompute the wavelengths
    #
    nfibers, npixmax = spplate_data['flux'].shape
    if 'align' in kwargs:
        loglam0 = allcoeff0[0] + allcoeff1[1]*np.arange(npixmax, dtype='d')
        spplate_data['loglam'] = np.resize(loglam0, (nfibers, npixmax))
    return spplate_data


def skymask(invvar, andmask, ormask=None, ngrow=2):
    """Mask regions where sky-subtraction errors are expected to dominate.

    Parameters
    ----------
    invvar : :class:`numpy.ndarray`
        Inverse variance.
    andmask : :class:`numpy.ndarray`
        An "and" mask.  For historical reasons, this input is ignored.
    ormask : :class:`numpy.ndarray`, optional
        An "or" mask.  Although technically this is optional, if it is
        not supplied, this function will have no effect.
    ngrow : :class:`int`, optional
        Expand bad areas by this number of pixels.

    Returns
    -------
    :class:`numpy.ndarray`
        The `invvar` multiplied by the bad areas.
    """
    from ..pydlutils.sdss import sdss_flagval
    from .. import smooth
    nrows, npix = invvar.shape
    badmask = np.zeros(invvar.shape, dtype='i4')
    badskychi = sdss_flagval('SPPIXMASK', 'BADSKYCHI')
    redmonster = sdss_flagval('SPPIXMASK', 'REDMONSTER')
    # brightsky = sdss_flagval('SPPIXMASK', 'BRIGHTSKY')
    if ormask is not None:
        badmask = badmask | ((ormask & badskychi) != 0)
        badmask = badmask | ((ormask & redmonster) != 0)
        # badmask = badmask | ((andmask & brightsky) != 0)
    if ngrow > 0:
        width = 2*ngrow + 1
        for k in range(nrows):
            badmask[k, :] = smooth(badmask[k, :]*width, width, True) > 0
    return invvar * (1 - badmask)


def spec_append(spec1, spec2, pixshift=0):
    """Append the array spec2 to the array spec1 & return a new array.

    If the dimension of these arrays is the same, then append as [spec1,spec2].
    If not, increase the size of the smaller array & fill with zeros.

    Parameters
    ----------
    spec1, spec2 : :class:`numpy.ndarray`
        Append `spec2` to `spec1`.
    pixshift : :class:`int`, optional
        If `pixshift` is set to a positive integer, `spec2` will be padded with
        `pixshift` zeros on the left side.  If `pixshift` is set to a
        negative integer, `spec1` will be padded with ``abs(pixshift)`` zeros
        on the left side.  If not set, all zeros will be padded on the right
        side.

    Returns
    -------
    :class:`numpy.ndarray`
        A new array containing both `spec1` and `spec2`.
    """
    nrows1, npix1 = spec1.shape
    nrows2, npix2 = spec2.shape
    nrows = nrows1+nrows2
    nadd1 = 0
    nadd2 = 0
    if pixshift != 0:
        if pixshift < 0:
            nadd1 = -pixshift
        else:
            nadd2 = pixshift
    maxpix = max(npix1 + nadd1, npix2 + nadd2)
    spec3 = np.zeros((nrows, maxpix), dtype=spec1.dtype)
    spec3[0:nrows1, nadd1:nadd1+npix1] = spec1
    spec3[nrows1:nrows, nadd2:nadd2+npix2] = spec2
    return spec3


def spec_path(plate, path=None, topdir=None, run2d=None):
    """Return the directory containing spPlate files.

    Parameters
    ----------
    plate : :class:`int` or :class:`numpy.ndarray`
        The plate(s) to examine.
    path : :class:`str`, optional
        If set, `path` becomes the full path for every plate. In other words,
        it completely short-circuits this function.
    topdir : :class:`str`, optional
        Used to override the value of :envvar:`BOSS_SPECTRO_REDUX`.
    run2d : :class:`str`, optional
        Used to override the value of :envvar:`RUN2D`.

    Returns
    -------
    :class:`list`
        A list of directories, one for each plate.

    Raises
    ------
    :exc:`KeyError`
        If environment variables are not supplied.
    """
    if isinstance(plate, (int,)) or plate.shape == ():
        platevec = np.array([plate], dtype='i4')
    else:
        platevec = plate
    if path is None:
        if run2d is None:
            run2d = os.environ['RUN2D']
        if topdir is None:
            env = "SPECTRO_REDUX"
            try:
                _ = int(run2d)
            except ValueError:
                env = 'BOSS_SPECTRO_REDUX'
            topdir = os.environ[env]
    paths = list()
    for p in platevec:
        if path is not None:
            paths.append(path)
        else:
            paths.append(os.path.join(topdir, run2d, '{0:04d}'.format(p)))
    return paths


def preprocess_spectra(flux, ivar, loglam=None, zfit=None, aesthetics='mean',
                       newloglam=None, wavemin=None, wavemax=None,
                       verbose=False):
    """Handle the processing of input spectra through the
    :func:`~pydl.pydlspec2d.spec2d.combine1fiber` stage.

    Parameters
    ----------
    flux : array-like
        The input spectral flux.
    ivar : array-like
        The inverse variance of the spectral flux.
    loglam : array-like, optional
        The input wavelength solution.
    zfit : array-like, optional
        The redshift of each input spectrum.
    aesthetics : :class:`str`, optional
        This parameter will be passed to
        :func:`~pydl.pydlspec2d.spec2d.combine1fiber`.
    newloglam : array-like, optional
        The output wavelength solution.
    wavemin : :class:`float`, optional
        Minimum wavelength if `newloglam` is not specified.
    wavemax : :class:`float`, optional
        Maximum wavelength if `newloglam` is not specified.
    verbose : :class:`bool`, optional
        If ``True``, print extra information.

    Returns
    -------
    :class:`tuple` of :class:`numpy.ndarray`
        The resampled flux, inverse variance and wavelength solution,
        respectively.
    """
    from .spec2d import combine1fiber
    if verbose:
        log.setLevel('DEBUG')
    if len(flux.shape) == 1:
        nobj = 1
        npix = flux.shape[0]
    else:
        nobj, npix = flux.shape
    #
    # The redshift of each object in pixels would be logshift/objdloglam.
    #
    if zfit is None:
        logshift = np.zeros((nobj,), dtype=flux.dtype)
    else:
        logshift = np.log10(1.0 + zfit)
    #
    # Determine the new wavelength mapping.
    #
    if loglam is None:
        if newloglam is None:
            raise ValueError("newloglam must be set if loglam is not!")
        return (flux, ivar, newloglam)
    else:
        if newloglam is None:
            igood = loglam != 0
            dloglam = loglam[1] - loglam[0]
            logmin = loglam[igood].min() - logshift.max()
            logmax = loglam[igood].max() - logshift.min()
            if wavemin is not None:
                logmin = max(logmin, np.log10(wavemin))
            if wavemax is not None:
                logmax = min(logmax, np.log10(wavemax))
            fullloglam = wavevector(logmin, logmax, binsz=dloglam)
        else:
            fullloglam = newloglam
            dloglam = fullloglam[1] - fullloglam[0]
        nnew = fullloglam.size
        fullflux = np.zeros((nobj, nnew), dtype='d')
        fullivar = np.zeros((nobj, nnew), dtype='d')
        #
        # Shift each spectrum to z = 0 and sample at the output wavelengths
        #
        if loglam.ndim == 1:
            indx = loglam > 0
            rowloglam = loglam[indx]
        for iobj in range(nobj):
            log.info("OBJECT %5d", iobj)
            if loglam.ndim > 1:
                if loglam.shape[0] != nobj:
                    raise ValueError('Wrong number of dimensions for loglam.')
                indx = loglam[iobj, :] > 0
                rowloglam = loglam[iobj, indx]
            flux1, ivar1 = combine1fiber(rowloglam-logshift[iobj],
                                         flux[iobj, indx], fullloglam,
                                         objivar=ivar[iobj, indx],
                                         binsz=dloglam, aesthetics=aesthetics,
                                         verbose=verbose)
            fullflux[iobj, :] = flux1
            fullivar[iobj, :] = ivar1
        return (fullflux, fullivar, fullloglam)


def template_metadata(inputfile, verbose=False):
    """Read template metadata from file.

    Parameters
    ----------
    inputfile : :class:`str`
        Name of a Parameter file containing the input data and metadata.
    verbose : :class:`bool`, optional
        If ``True``, print lots of extra information.

    Returns
    -------
    :class:`tuple`
        A tuple containing the list of input spectra and a dictionary
        containing other metadata.
    """
    from ..pydlutils.yanny import yanny
    if verbose:
        log.setLevel('DEBUG')
    if not os.path.exists(inputfile):
        raise Pydlspec2dException("Could not find {0}!".format(inputfile))
    log.debug("Reading input data from %s.", inputfile)
    par = yanny(inputfile)
    required_metadata = {'object': str, 'method': str, 'aesthetics': str,
                         'run2d': str, 'run1d': str,
                         'wavemin': float, 'wavemax': float, 'snmax': float,
                         'niter': int, 'nkeep': int, 'minuse': int}
    metadata = dict()
    for key in required_metadata:
        try:
            metadata[key] = required_metadata[key](par[key])
            log.debug('%s = %s', key, par[key])
        except KeyError:
            raise KeyError('The {0} keyword was not found in {1}!'.format(key, inputfile))
        except ValueError:
            raise ValueError('The {0} keyword has invalid value, {1}!'.format(key, par[key]))
    slist = par['EIGENOBJ']
    for r in ('run2d', 'run1d'):
        try:
            metadata['orig_'+r] = os.environ[r.upper()]
        except KeyError:
            metadata['orig_'+r] = None
        os.environ[r.upper()] = metadata[r]
    if metadata['method'].lower() == 'hmf':
        required_hmf_metadata = {'nonnegative': lambda x: bool(int(x)),
                                 'epsilon': float}
        for key in required_hmf_metadata:
            try:
                metadata[key] = required_hmf_metadata[key](par[key])
            except KeyError:
                raise KeyError('The {0} keyword was not found in {1}!'.format(key, inputfile))
            except ValueError:
                raise ValueError('The {0} keyword has invalid value, {1}!'.format(key, par[key]))
    return (slist, metadata)


def template_input(inputfile, dumpfile, flux=False, verbose=False):
    """Collect spectra and pass them to PCA or HMF solvers to compute
    spectral templates.

    This function replaces the various ``PCA_GAL()``, ``PCA_STAR()``, etc.,
    functions from idlspec2d.

    Parameters
    ----------
    inputfile : :class:`str`
        Name of a Parameter file containing the input data and metadata.
    dumpfile : :class:`str`
        Name of a Pickle file used to store intermediate data.
    flux : :class:`bool`, optional
        If ``True``, plot the individual input spectra.
    verbose : :class:`bool`, optional
        If ``True``, print lots of extra information.
    """
    import pickle
    from astropy.constants import c as cspeed
    from .. import __version__ as pydl_version
    from ..goddard.astro import get_juldate
    from ..pydlutils.image import djs_maskinterp
    from ..pydlutils.math import djs_median
    #
    # Logging
    #
    if verbose:
        log.setLevel('DEBUG')
    #
    # Read metadata.
    #
    slist, metadata = template_metadata(inputfile)
    #
    # Name the output files.
    #
    jd = get_juldate()
    outfile = "spEigen{0}-{1:d}".format(metadata['object'].title(), int(jd - 2400000.5))
    #
    # Read the input spectra
    #
    if os.path.exists(dumpfile):
        log.info("Loading data from %s.", dumpfile)
        with open(dumpfile) as f:
            inputflux = pickle.load(f)
        newflux = inputflux['newflux']
        newivar = inputflux['newivar']
        newloglam = inputflux['newloglam']
    else:
        if metadata['object'].lower() == 'star':
            spplate = readspec(slist.plate, mjd=slist.mjd, fiber=slist.fiberid,
                               align=True)
        else:
            spplate = readspec(slist.plate, mjd=slist.mjd, fiber=slist.fiberid)
        #
        # Insist that all of the requested spectra exist.
        #
        missing = spplate['plugmap']['FIBERID'] == 0
        if missing.any():
            imissing = missing.nonzero()[0]
            for k in imissing:
                log.error("Missing plate=%d mjd=%d fiberid=%d",
                          slist.plate[k], slist.mjd[k], slist.fiberid[k])
            raise ValueError("{0:d} missing object(s).".format(missing.sum()))
        #
        # Do not fit where the spectrum may be dominated by sky-sub residuals.
        #
        objinvvar = skymask(spplate['invvar'], spplate['andmask'],
                            spplate['ormask'])
        ifix = spplate['flux']**2 * objinvvar > metadata['snmax']**2
        if ifix.any():
            objinvvar[ifix.nonzero()] = (metadata['snmax']/spplate['flux'][ifix.nonzero()])**2
        #
        # Set the new wavelength mapping here.  If the binsz keyword is not set,
        # then bin size is determined from the first spectrum returned by readspec.
        # This is fine in the case where all spectra have the same bin size
        # (though their starting wavelengths may differ).  However, this may not
        # be a safe assumption in the future.
        #
        try:
            objdloglam = float(metadata['binsz'])
        except (KeyError, ValueError):
            objdloglam = spplate['loglam'][0, 1] - spplate['loglam'][0, 0]
        if metadata['object'].lower() == 'star':
            newloglam = spplate['loglam'][0, :]
        else:
            newloglam = wavevector(np.log10(metadata['wavemin']),
                                   np.log10(metadata['wavemax']), binsz=objdloglam)
        try:
            zfit = slist.zfit
        except AttributeError:
            zfit = slist.cz/cspeed.to('km / s').value
        #
        # Shift to common wavelength grid.
        #
        newflux, newivar, newloglam = preprocess_spectra(spplate['flux'],
                                                         objinvvar,
                                                         loglam=spplate['loglam'],
                                                         zfit=zfit,
                                                         newloglam=newloglam,
                                                         aesthetics=metadata['aesthetics'],
                                                         verbose=verbose)
        #
        # Dump input fluxes to a file for debugging purposes.
        #
        if not os.path.exists(dumpfile):
            with open(dumpfile, 'w') as f:
                inputflux = {'newflux': newflux, 'newivar': newivar,
                             'newloglam': newloglam}
                pickle.dump(inputflux, f)
    #
    # Solve.
    #
    if metadata['object'].lower() == 'qso':
        pcaflux = template_qso(metadata, newflux, newivar, verbose)
    elif metadata['object'].lower() == 'star':
        pcaflux = template_star(metadata, newloglam, newflux, newivar,
                                slist, outfile, verbose)
    else:
        if metadata['method'].lower() == 'pca':
            pcaflux = pca_solve(newflux, newivar,
                                niter=metadata['niter'],
                                nkeep=metadata['nkeep'],
                                verbose=verbose)
        elif metadata['method'].lower() == 'hmf':
            hmf = HMF(newflux, newivar,
                      K=metadata['nkeep'],
                      n_iter=metadata['niter'],
                      nonnegative=metadata['nonnegative'],
                      epsilon=metadata['epsilon'],
                      verbose=verbose)
            pcaflux = hmf.solve()
        else:
            raise ValueError("Unknown method: {0}!".format(metadata['method']))
    pcaflux['newflux'] = newflux
    pcaflux['newivar'] = newivar
    pcaflux['newloglam'] = newloglam
    #
    # Fill in bad data with a running median of the good data.
    # The presence of boundary='nearest' means that this code snippet
    # was never meant to be called!  In other words it should always
    # be the case that qgood.all() is True.
    #
    if 'usemask' in pcaflux:
        qgood = pcaflux['usemask'] >= metadata['minuse']
        if not qgood.all():
            warn("Would have triggered djs_median replacement!", Pydlspec2dUserWarning)
        if False:
            medflux = np.zeros(pcaflux['flux'].shape, dtype=pcaflux['flux'].dtype)
            for i in range(metadata['nkeep']):
                medflux[i, qgood] = djs_median(pcaflux['flux'][i, qgood],
                                               width=51, boundary='nearest')
                medflux[i, :] = djs_maskinterp(medflux[i, :], ~qgood, const=True)
            pcaflux['flux'][:, ~qgood] = medflux[:, ~qgood]
    #
    # Make plots
    #
    colorvec = ['k', 'r', 'g', 'b', 'm', 'c']
    smallfont = FontProperties(size='xx-small')
    nspectra = pcaflux['newflux'].shape[0]
    #
    # Plot input spectra
    #
    if flux:
        nfluxes = 30
        separation = 5.0
        nplots = nspectra/nfluxes
        if nspectra % nfluxes > 0:
            nplots += 1
        for k in range(nplots):
            istart = k*nfluxes
            iend = min(istart+nfluxes, nspectra) - 1
            fig = plt.figure(dpi=100)
            ax = fig.add_subplot(111)
            for l in range(istart, iend+1):
                _ = ax.plot(10.0**pcaflux['newloglam'],
                            pcaflux['newflux'][l, :] + separation*(l % nfluxes),
                            colorvec[l % len(colorvec)]+'-',
                            linewidth=1)
            ax.set_xlabel(r'Wavelength [$\AA$]')
            ax.set_ylabel(r'Flux [$\mathsf{10^{-17} erg\, cm^{-2} s^{-1} \AA^{-1}}$] + Constant')
            ax.set_title('Input Spectra {0:04d}-{1:04d}'.format(istart+1, iend+1))
            ax.set_ylim(pcaflux['newflux'][istart, :].min(), pcaflux['newflux'][iend-1, :].max()+separation*(nfluxes-1))
            fig.savefig('{0}.flux.{1:04d}-{2:04d}.png'.format(outfile, istart+1, iend+1))
            plt.close(fig)
    #
    # Missing data diagnostic.
    #
    fig = plt.figure(dpi=100)
    ax = fig.add_subplot(111)
    _ = ax.plot(10.0**pcaflux['newloglam'], (pcaflux['newivar'] == 0).sum(0)/float(nspectra), 'k-')
    ax.set_xlabel(r'Wavelength [$\AA$]')
    ax.set_ylabel('Fraction of spectra with missing data')
    ax.set_title('Missing Data')
    ax.grid(True)
    fig.savefig(outfile+'.missing.png')
    plt.close(fig)
    #
    # usemask diagnostic
    #
    if 'usemask' in pcaflux:
        fig = plt.figure(dpi=100)
        ax = fig.add_subplot(111)
        _ = ax.semilogy(10.0**pcaflux['newloglam'][pcaflux['usemask'] > 0],
                        pcaflux['usemask'][pcaflux['usemask'] > 0], 'k-',
                        10.0**pcaflux['newloglam'],
                        np.zeros(pcaflux['newloglam'].shape,
                        dtype=pcaflux['newloglam'].dtype) + metadata['minuse'],
                        'k--')
        ax.set_xlabel(r'Wavelength [$\AA$]')
        ax.set_ylabel('Usemask')
        ax.set_title('UseMask')
        ax.grid(True)
        fig.savefig(outfile+'.usemask.png')
        plt.close(fig)
    #
    # This type of figure isn't really meaningful for stars.
    #
    if metadata['object'].lower() != 'star':
        aratio10 = pcaflux['acoeff'][:, 1]/pcaflux['acoeff'][:, 0]
        aratio20 = pcaflux['acoeff'][:, 2]/pcaflux['acoeff'][:, 0]
        aratio30 = pcaflux['acoeff'][:, 3]/pcaflux['acoeff'][:, 0]
        fig = plt.figure(dpi=100)
        ax = fig.add_subplot(111)
        _ = ax.plot(aratio10, aratio20, marker='None', linestyle='None')
        for k in range(len(aratio10)):
            _ = ax.text(aratio10[k], aratio20[k],
                        '{0:04d}-{1:04d}'.format(slist.plate[k], slist.fiberid[k]),
                        horizontalalignment='center', verticalalignment='center',
                        color=colorvec[k % len(colorvec)],
                        fontproperties=smallfont)
        # ax.set_xlim([aratio10.min(), aratio10.max])
        # ax.set_xlim([aratio20.min(), aratio20.max])
        ax.set_xlabel('Eigenvalue Ratio, $a_1/a_0$')
        ax.set_ylabel('Eigenvalue Ratio, $a_2/a_0$')
        ax.set_title('Eigenvalue Ratios')
        fig.savefig(outfile+'.a2_v_a1.png')
        plt.close(fig)
        fig = plt.figure(dpi=100)
        ax = fig.add_subplot(111)
        _ = ax.plot(aratio20, aratio30, marker='None', linestyle='None')
        for k in range(len(aratio10)):
            _ = ax.text(aratio20[k], aratio30[k],
                        '{0:04d}-{1:04d}'.format(slist.plate[k], slist.fiberid[k]),
                        horizontalalignment='center', verticalalignment='center',
                        color=colorvec[k % len(colorvec)],
                        fontproperties=smallfont)
        # ax.set_xlim([aratio10.min(), aratio10.max])
        # ax.set_xlim([aratio20.min(), aratio20.max])
        ax.set_xlabel('Eigenvalue Ratio, $a_2/a_0$')
        ax.set_ylabel('Eigenvalue Ratio, $a_3/a_0$')
        ax.set_title('Eigenvalue Ratios')
        fig.savefig(outfile+'.a3_v_a2.png')
        plt.close(fig)
    #
    # Save output to FITS file.
    #
    if os.path.exists(outfile+'.fits'):
        os.remove(outfile+'.fits')
    hdu0 = fits.PrimaryHDU(pcaflux['flux'])
    objtypes = {'gal': 'GALAXY', 'qso': 'QSO', 'star': 'STAR'}
    if not pydl_version:
        pydl_version = 'git'
    hdu0.header['OBJECT'] = (objtypes[metadata['object']], 'Type of template')
    hdu0.header['COEFF0'] = (pcaflux['newloglam'][0], 'Wavelength zeropoint')
    hdu0.header['COEFF1'] = (pcaflux['newloglam'][1]-pcaflux['newloglam'][0], 'Delta wavelength')
    hdu0.header['IDLUTILS'] = ('pydl-{0}'.format(pydl_version), 'Version of idlutils')
    hdu0.header['SPEC2D'] = ('pydl-{0}'.format(pydl_version), 'Version of idlspec2d')
    hdu0.header['RUN2D'] = (os.environ['RUN2D'], 'Version of 2d reduction')
    hdu0.header['RUN1D'] = (os.environ['RUN1D'], 'Version of 1d reduction')
    hdu0.header['FILENAME'] = (inputfile, 'Input file')
    hdu0.header['METHOD'] = (metadata['method'].upper(), 'Method used')
    if metadata['method'].lower() == 'hmf':
        hdu0.header['NONNEG'] = (metadata['nonnegative'], 'Was nonnegative HMF used?')
        hdu0.header['EPSILON'] = (metadata['epsilon'], 'Regularization parameter used.')
    # for i in range(len(namearr)):
    #     hdu0.header["NAME{0:d}".format(i)] = namearr[i]+' '
    c = [fits.Column(name='plate', format='J', array=slist.plate),
         fits.Column(name='mjd', format='J', array=slist.mjd),
         fits.Column(name='fiberid', format='J', array=slist.fiberid)]
    if metadata['object'].lower() == 'star':
        c.append(fits.Column(name='cz', format='D', unit='km/s',
                 array=slist.cz))
        for i, name in enumerate(pcaflux['namearr']):
            hdu0.header['NAME{0:d}'.format(i)] = (name, 'Name of class {0:d}.'.format(i))
    else:
        c.append(fits.Column(name='zfit', format='D', array=slist.zfit))
    hdu1 = fits.BinTableHDU.from_columns(fits.ColDefs(c))
    hdulist = fits.HDUList([hdu0, hdu1])
    hdulist.writeto(outfile+'.fits')
    if metadata['object'].lower() != 'star':
        plot_eig(outfile+'.fits')
    #
    # Clean up
    #
    for r in ('run2d', 'run1d'):
        if metadata['orig_'+r] is None:
            del os.environ[r.upper()]
        else:
            os.environ[r.upper()] = metadata['orig_'+r]
    return


def template_qso(metadata, newflux, newivar, verbose=False):
    """Run PCA or HMF on QSO spectra.

    Historically, QSO templates were comptuted one at a time instead of
    all at once.

    Parameters
    ----------
    metadata : :class:`dict`
        Dictionary containing metadata about the spectra.
    newflux : :class:`~numpy.ndarray`
        Flux shifted onto common wavelength.
    newivar : :class:`~numpy.ndarray`
        Inverse variances of the fluxes.
    verbose : :class:`bool`, optional
        If ``True``, print lots of extra information.

    Returns
    -------
    :class:`dict`
        A dictonary containing flux, eigenvalues, etc.
    """
    from ..pydlutils.math import computechi2
    if metadata['object'].lower() != 'qso':
        raise Pydlspec2dException("You appear to be passing the wrong kind of object to template_qso()!")
    if len(newflux.shape) == 1:
        nobj = 1
        npix = newflux.shape[0]
    else:
        nobj, npix = newflux.shape
    objflux = newflux.copy()
    for ikeep in range(metadata['nkeep']):
        log.info("Solving for eigencomponent #%d of %d", ikeep+1, metadata['nkeep'])
        if metadata['method'].lower() == 'pca':
            pcaflux1 = pca_solve(objflux, newivar,
                                 niter=metadata['niter'], nkeep=1,
                                 verbose=verbose)
        elif metadata['method'].lower() == 'hmf':
            hmf = HMF(objflux, newivar,
                      K=metadata['nkeep'],
                      n_iter=metadata['niter'],
                      nonnegative=metadata['nonnegative'],
                      epsilon=metadata['epsilon'],
                      verbose=verbose)
            pcaflux1 = hmf.solve()
        else:
            raise ValueError("Unknown method: {0}!".format(metadata['method']))
        if ikeep == 0:
            #
            # Create new pcaflux dict
            #
            pcaflux = dict()
            for k in pcaflux1:
                pcaflux[k] = pcaflux1[k].copy()
        else:
            #
            # Add to existing dict
            #
            # for k in pcaflux1:
            #     pcaflux[k] = np.vstack((pcaflux[k],pcaflux1[k]))
            pcaflux['flux'] = np.vstack((pcaflux['flux'], pcaflux1['flux']))
            pcaflux['eigenval'] = np.concatenate((pcaflux['eigenval'], pcaflux1['eigenval']))
            #
            # Re-solve for the coefficients using all PCA components so far
            #
            acoeff = np.zeros((nobj, ikeep+1), dtype=pcaflux1['acoeff'].dtype)
            for iobj in range(nobj):
                out = computechi2(newflux[iobj, :],
                                  np.sqrt(pcaflux1['newivar'][iobj, :]),
                                  pcaflux['flux'].T)
                acoeff[iobj, :] = out['acoeff']
            #
            # Prevent re-binning of spectra on subsequent calls to pca_solve()
            #
            # objloglam = None
            if ikeep == 0:
                objflux = newflux - np.outer(acoeff, pcaflux['flux'])
            else:
                objflux = newflux - np.dot(acoeff, pcaflux['flux'])
            # objflux = newflux - np.outer(acoeff,pcaflux['flux'])
            # objinvvar = pcaflux1['newivar']
            pcaflux['acoeff'] = acoeff
    return pcaflux


def template_star(metadata, newloglam, newflux, newivar, slist, outfile,
                  verbose=False):
    """Run PCA or HMF on stellar spectra of various classes.

    Parameters
    ----------
    metadata : :class:`dict`
        Dictionary containing metadata about the spectra.
    newloglam : :class:`~numpy.ndarray`
        The wavelength array, used only for plots.
    newflux : :class:`~numpy.ndarray`
        Flux shifted onto common wavelength.
    newivar : :class:`~numpy.ndarray`
        Inverse variances of the fluxes.
    slist : :class:`~numpy.recarray`
        The list of objects, containing stellar class information.
    outfile : :class:`str`
        The base name of output file, used for plots.
    verbose : :class:`bool`, optional
        If ``True``, print lots of extra information.

    Returns
    -------
    :class:`dict`
        A dictonary containing flux, eigenvalues, etc.
    """
    from .. import uniq
    from ..pydlutils.image import djs_maskinterp
    if metadata['object'].lower() != 'star':
        raise Pydlspec2dException("You appear to be passing the wrong kind of object to template_star()!")
    #
    # Find the list of unique star types
    #
    isort = np.argsort(slist['class'])
    classlist = slist['class'][isort[uniq(slist['class'][isort])]]
    #
    # Loop over each star type
    #
    npix, nstars = newflux.shape
    pcaflux = dict()
    pcaflux['namearr'] = list()
    for c in classlist:
        #
        # Find the subclasses for this stellar type
        #
        log.info("Finding eigenspectra for Stellar class %s.", c)
        indx = (slist['class'] == c).nonzero()[0]
        nindx = indx.size
        thesesubclass = slist['subclass'][indx]
        isort = np.argsort(thesesubclass)
        subclasslist = thesesubclass[isort[uniq(thesesubclass[isort])]]
        nsubclass = subclasslist.size
        #
        # Solve for 2 eigencomponents if we have specified subclasses for
        # this stellar type
        #
        if nsubclass == 1:
            nkeep = 1
        else:
            nkeep = 2
        if metadata['method'].lower() == 'pca':
            pcaflux1 = pca_solve(newflux[indx, :], newivar[indx, :],
                                 niter=metadata['niter'], nkeep=nkeep,
                                 verbose=verbose)
        elif metadata['method'].lower() == 'hmf':
            hmf = HMF(newflux[indx, :], newivar[indx, :],
                      K=metadata['nkeep'],
                      n_iter=metadata['niter'],
                      nonnegative=metadata['nonnegative'],
                      epsilon=metadata['epsilon'],
                      verbose=verbose)
            pcaflux1 = hmf.solve()
        else:
            raise ValueError("Unknown method: {0}!".format(metadata['method']))
        #
        # Some star templates are generated from only one spectrum,
        # and these will not have a usemask set.
        #
        if 'usemask' not in pcaflux1:
            pcaflux1['usemask'] = np.zeros((npix,), dtype='i4') + nindx
        #
        # Interpolate over bad flux values in the middle of a spectrum,
        # and set fluxes to zero at the blue+red ends of the spectrum
        #
        # minuse = 1 # ?
        minuse = np.floor((nindx+1) / 3.0)
        qbad = pcaflux1['usemask'] < minuse
        #
        # Interpolate over all bad pixels
        #
        for j in range(nkeep):
            pcaflux1['flux'][j, :] = djs_maskinterp(pcaflux1['flux'][j, :],
                                                    qbad, const=True)
        #
        # Set bad pixels at the very start or end of the spectrum to zero
        # instead.
        #
        npix = qbad.size
        igood = (~qbad).nonzero()[0]
        if qbad[0]:
            pcaflux1['flux'][:, 0:igood[0]-1] = 0
        if qbad[npix-1]:
            pcaflux1['flux'][:, igood[::-1][0]+1:npix] = 0
        #
        # Re-normalize the first eigenspectrum to a mean of 1
        #
        norm = pcaflux1['flux'][0, :].mean()
        pcaflux1['flux'] /= norm
        if 'acoeff' in pcaflux1:
            pcaflux1['acoeff'] *= norm
        #
        # Now loop through each stellar subclass and reconstruct
        # an eigenspectrum for that subclass
        #
        thesesubclassnum = np.zeros(thesesubclass.size, dtype='i4')
        colorvec = ['k', 'r', 'g', 'b', 'm', 'c']
        smallfont = FontProperties(size='xx-small')
        fig = plt.figure(dpi=100)
        ax = fig.add_subplot(111)
        for isub in range(nsubclass):
            ii = (thesesubclass == subclasslist[isub]).nonzero()[0]
            thesesubclassnum[ii] = isub
            if nkeep == 1:
                thisflux = pcaflux1['flux'][0, :]
            else:
                aratio = pcaflux1['acoeff'][ii, 1]/pcaflux1['acoeff'][ii, 0]
                #
                # np.median(foo) is equivalent to MEDIAN(foo,/EVEN)
                #
                thisratio = np.median(aratio)
                thisflux = (pcaflux1['flux'][0, :] +
                            thisratio.astype('f') * pcaflux1['flux'][1, :])
            #
            # Plot spectra
            #
            plotflux = thisflux/thisflux.max()
            ax.plot(10.0**newloglam, plotflux,
                    "{0}-".format(colorvec[isub % len(colorvec)]),
                    linewidth=1)
            if isub == 0:
                ax.set_xlabel(r'Wavelength [$\AA$]')
                ax.set_ylabel('Flux [arbitrary units]')
                ax.set_title('STAR {0}: Eigenspectra Reconstructions'.format(c))
            _ = ax.text(10.0**newloglam[-1], plotflux[-1],
                        subclasslist[isub],
                        horizontalalignment='right', verticalalignment='center',
                        color=colorvec[isub % len(colorvec)], fontproperties=smallfont)
            fig.savefig(outfile+'.{0}.png'.format(c))
            plt.close(fig)
            if 'flux' in pcaflux:
                pcaflux['flux'] = np.vstack((pcaflux['flux'], thisflux))
                # pcaflux['acoeff'] = np.vstack((pcaflux['acoeff'], pcaflux1['acoeff']))
                # pcaflux['usemask'] = np.vstack((pcaflux['usemask'], pcaflux1['usemask']))
            else:
                pcaflux['flux'] = thisflux
                # pcaflux['acoeff'] = pcaflux1['acoeff']
                # pcaflux['usemask'] = pcaflux1['usemask']
            pcaflux['namearr'].append(subclasslist[isub])
    return pcaflux


def template_input_main():  # pragma: no cover
    """Entry point for the compute_templates script.

    Returns
    -------
    :class:`int`
        An integer suitable for passing to :func:`sys.exit`.
    """
    #
    # Imports for main()
    #
    import sys
    from argparse import ArgumentParser

    # Get home directory in platform-independent way
    home_dir = os.path.expanduser('~')
    #
    # Get Options
    #
    parser = ArgumentParser(description="Compute spectral templates.",
                            prog=os.path.basename(sys.argv[0]))
    parser.add_argument('-d', '--dump', action='store', dest='dump',
                        metavar='FILE',
                        default=os.path.join(home_dir, 'scratch', 'templates', 'compute_templates.dump'),
                        help='Dump data to a pickle file (default: %(default)s).')
    parser.add_argument('-F', '--flux', action='store_true', dest='flux',
                        help='Plot input spectra.')
    parser.add_argument('-f', '--file', action='store', dest='inputfile',
                        metavar='FILE',
                        default=os.path.join(home_dir, 'scratch', 'templates', 'compute_templates.par'),
                        help='Read input spectra and redshifts from FILE (default: %(default)s).')
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help='Print lots of extra information.')
    options = parser.parse_args()
    template_input(options.inputfile, options.dump, options.flux, options.verbose)
    return 0


def wavevector(minfullwave, maxfullwave, zeropoint=3.5, binsz=1.0e-4,
               wavemin=None):
    """Return an array of wavelengths.

    Parameters
    ----------
    minfullwave : :class:`float`
        Minimum wavelength.
    maxfullwave : :class:`float`
        Maximum wavelength.
    zeropoint : :class:`float`, optional
        Offset of the input wavelength values.
    binsz : :class:`float`, optional
        Separation between wavelength values.
    wavemin : :class:`float`, optional
        If this is set the values of `minfullwave` and `zeropoint` are ignored.

    Returns
    -------
    :class:`numpy.ndarray`
        Depending on the values of `minfullwave`, `binsz`, etc., the resulting
        array could be interpreted as an array of wavelengths or an array of
        log(wavelength).
    """
    if wavemin is not None:
        spotmin = 0
        spotmax = int((maxfullwave - wavemin)/binsz)
        # wavemax = spotmax * binsz + wavemin
    else:
        spotmin = int((minfullwave - zeropoint)/binsz) + 1
        spotmax = int((maxfullwave - zeropoint)/binsz)
        wavemin = spotmin * binsz + zeropoint
        # wavemax = spotmax * binsz + zeropoint
    nfinalpix = spotmax - spotmin + 1
    finalwave = np.arange(nfinalpix, dtype='d') * binsz + wavemin
    return finalwave



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import pytest
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from .. import PydlutilsException, PydlutilsUserWarning
from .. import mangle as mng


class TestMangle(object):
    """Test the functions in pydl.pydlutils.mangle.
    """

    def setup_method(self):
        self.poly_fits = get_pkg_data_filename('t/polygon.fits')
        self.no_id_fits = get_pkg_data_filename('t/polygon_no_id.fits')
        self.one_cap_fits = get_pkg_data_filename('t/polygon_one_cap.fits')
        self.poly_ply = get_pkg_data_filename('t/polygon.ply')
        self.bad_ply = get_pkg_data_filename('t/median_data.txt')

    def test_ManglePolygon(self):
        #
        # Zero caps
        #
        poly = mng.ManglePolygon()
        assert np.allclose(poly.str, 4.0*np.pi)
        assert poly.cmminf() is None
        assert not poly.gzeroar()
        assert np.allclose(poly.garea(), 4.0*np.pi)
        #
        # One cap.
        #
        x = np.array([[0.0, 0.0, 1.0]])
        cm = np.array([0.5])
        poly = mng.ManglePolygon(x=x, cm=cm)
        assert np.allclose(poly.garea(), np.pi)
        #
        # Bad inputs
        #
        with pytest.raises(ValueError):
            poly = mng.ManglePolygon(weight=1.0)
        #
        # Multiple caps
        #
        x = np.array([[0.0, 0.0, 1.0],
                      [1.0, 0.0, 0.0],
                      [0.0, 1.0, 0.0]])
        cm = np.array([1.0, 1.0, 1.0])
        poly = mng.ManglePolygon(x=x, cm=cm, str=np.pi/2.0)
        assert poly.ncaps == 3
        assert poly.weight == 1.0
        assert poly.use_caps == (1 << 3) - 1
        poly = mng.ManglePolygon(x=x, cm=cm, weight=0.5)
        assert poly.weight == 0.5
        poly = mng.ManglePolygon(x=x, cm=cm, pixel=20)
        assert poly.pixel == 20
        poly = mng.ManglePolygon(x=x, cm=cm, use_caps=3, str=np.pi/2.0)
        assert poly.use_caps == 3
        poly2 = poly.copy()
        assert poly2.use_caps == poly.use_caps
        assert (poly2.cm == poly.cm).all()
        assert np.allclose(poly2.str, np.pi/2.0)
        x = np.array([[0.0, 0.0, 1.0],
                      [1.0, 0.0, 0.0]])
        cm = np.array([1.0, 1.0])
        poly = mng.ManglePolygon(x=x, cm=cm)
        poly2 = poly.add_caps(np.array([[0.0, 1.0, 0.0], ]), np.array([1.0, ]))
        assert poly2.ncaps == 3
        assert poly2.use_caps == poly.use_caps
        with pytest.warns(PydlutilsUserWarning):
            assert poly2.str == 1.0  # dummy value!
        poly3 = poly.polyn(poly2, 2)
        assert poly3.ncaps == 3
        assert poly3.use_caps == poly.use_caps
        assert np.allclose(poly3.x[2, :], np.array([0.0, 1.0, 0.0]))
        poly3 = poly.polyn(poly2, 2, complement=True)
        assert poly3.ncaps == 3
        assert poly3.use_caps == poly.use_caps
        assert np.allclose(poly3.cm[2], -1.0)

    def test_angles_to_x(self):
        x = mng.angles_to_x(np.array([[0.0, 0.0], [90.0, 90.0],
                                      [0.0, 90.0]]))
        assert np.allclose(x, np.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0],
                                        [1.0, 0.0, 0.0]]))
        a = mng.angles_to_x(np.array([[0.0, 90.0], [90.0, 0.0],
                                      [0.0, 0.0]]), latitude=True)
        assert np.allclose(a, np.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0],
                                        [1.0, 0.0, 0.0]]))

    def test_cap_distance(self):
        x = np.array([0.0, 0.0, 1.0])
        cm = 1.0
        with pytest.raises(ValueError):
            d = mng.cap_distance(x, cm, np.array([[1.0, 2.0, 3.0, 4.0], ]))
        d = mng.cap_distance(x, cm, np.array([[0.0, 45.0], ]))
        assert np.allclose(d, np.array([45.0]))
        y = mng.angles_to_x(np.array([[0.0, 45.0], ]), latitude=True)
        d = mng.cap_distance(x, cm, y)
        assert np.allclose(d, np.array([45.0]))
        d = mng.cap_distance(x, cm, np.array([[0.0, -45.0], ]))
        assert np.allclose(d, np.array([-45.0]))
        d = mng.cap_distance(x, -1.0, np.array([[0.0, -45.0], ]))
        assert np.allclose(d, np.array([45.0]))

    def test_circle_cap(self):
        with pytest.raises(ValueError):
            x, cm = mng.circle_cap(90.0, np.array([[1.0, 2.0, 3.0, 4.0], ]))
        xin = np.array([[0.0, 0.0, 1.0], ])
        x, cm = mng.circle_cap(90.0, xin)
        assert np.allclose(x, xin)
        assert np.allclose(cm, 1.0)
        radec = np.array([[0.0, 90.0], ])
        x, cm = mng.circle_cap(90.0, radec)
        assert np.allclose(x, xin)
        assert np.allclose(cm, 1.0)
        x, cm = mng.circle_cap(np.float32(90.0), radec)
        assert np.allclose(x, xin)
        assert np.allclose(cm, 1.0)
        x, cm = mng.circle_cap(np.array([90.0, ]), radec)
        assert np.allclose(x, xin)
        assert np.allclose(cm, np.array([1.0, ]))
        with pytest.raises(ValueError):
            x, cm = mng.circle_cap(np.array([90.0, 90.0]), radec)

    def test_is_cap_used(self):
        assert mng.is_cap_used(1 << 2, 2)
        assert not mng.is_cap_used(1 << 2, 1)

    def test_is_in_cap(self):
        x = np.array([0.0, 0.0, 1.0])
        cm = 1.0
        d = mng.is_in_cap(x, cm, np.array([[0.0, 45.0], [0.0, -45.0]]))
        assert (d == np.array([True, False])).all()

    def test_is_in_polygon(self):
        x = np.array([[0.0, 0.0, 1.0],
                      [0.0, 1.0, 0.0],
                      [1.0, 0.0, 0.0]])
        cm = np.array([1.0, 1.0, 1.0])
        p = mng.ManglePolygon(x=x, cm=cm)
        d = mng.is_in_polygon(p, np.array([[45.0, 45.0], [135.0, -45.0]]))
        assert (d == np.array([True, False])).all()
        d = mng.is_in_polygon(p, np.array([[45.0, 45.0], [-45.0, 45.0]]),
                              ncaps=2)
        assert (d == np.array([True, False])).all()
        poly = mng.read_fits_polygons(self.no_id_fits)
        d = mng.is_in_polygon(poly[0],
                              np.array([[0.0, 10.0],
                                        [4.0, 4.5],
                                        [90, 4.0],
                                        [135, 3.0],
                                        [180, 0.5],
                                        [270, 0.25]]))
        assert (d == np.array([False, True, False, False, False, False])).all()

    def test_is_in_window(self):
        np.random.seed(271828)
        RA = 7.0*np.random.random(1000) + 268.0
        Dec = 90.0 - np.degrees(np.arccos(0.08*np.random.random(1000)))
        points = np.vstack((RA, Dec)).T
        poly = mng.read_fits_polygons(self.poly_fits)
        i = mng.is_in_window(poly, points)
        assert i[0].sum() == 3

    def test_read_fits_polygons(self):
        poly = mng.read_fits_polygons(self.poly_fits)
        use_caps = np.array([31, 15, 31, 7, 31, 15, 15, 7, 15, 15,
                             15, 31, 15, 15, 15, 15, 15, 15, 31, 15],
                            dtype=np.uint32)
        #
        # Attribute access doesn't work on unsigned columns.
        #
        assert (poly['USE_CAPS'] == use_caps).all()
        assert (poly['use_caps'] == use_caps).all()
        with pytest.raises(AttributeError):
            _ = poly.no_such_attribute
        cm0 = np.array([-1.0, -0.99369437, 1.0, -1.0, 0.00961538])
        assert np.allclose(poly.cm[0][0:poly.ncaps[0]], cm0)
        assert poly[0]['NCAPS'] == 5
        poly = mng.read_fits_polygons(self.poly_fits, convert=True)
        assert poly[0].use_caps == 31
        assert np.allclose(poly[0].cm, cm0)
        assert poly[0].cmminf() == 4
        #
        # A FITS file might not contain IFIELD.
        #
        poly = mng.read_fits_polygons(self.no_id_fits)
        assert len(poly) == 1
        #
        # A FITS file might contain exactly one polygon with exactly one cap.
        #
        poly = mng.read_fits_polygons(self.one_cap_fits, convert=True)
        assert poly[0].ncaps == 1

    def test_read_mangle_polygons(self):
        with pytest.raises(PydlutilsException):
            poly = mng.read_mangle_polygons(self.bad_ply)
        poly = mng.read_mangle_polygons(self.poly_ply)
        assert len(poly.header) == 3
        assert poly.header[0] == 'pixelization 6s'
        assert len(poly) == 4
        assert np.allclose(poly[0].x[0, :],
                           np.array([0.0436193873653360, 0.9990482215818578,
                                     0.0]))
        assert poly[3].ncaps == 3

    def test_set_use_caps(self):
        poly = mng.read_fits_polygons(self.poly_fits, convert=True)
        # old_use_caps = poly[0].use_caps
        index_list = list(range(poly[0].ncaps))
        use_caps = mng.set_use_caps(poly[0], index_list, allow_doubles=True)
        assert use_caps == poly[0].use_caps
        use_caps = mng.set_use_caps(poly[0], index_list)
        assert use_caps == poly[0].use_caps
        x = np.array([[0.0, 0.0, 1.0],
                      [0.0, 1.0, 0.0],
                      [1.0, 0.0, 0.0],
                      [1.0-1.0e-8, 1.0e-8, 0.0]])
        cm = np.array([1.0, 1.0, 1.0, 1.0-1.0e-8])
        p = mng.ManglePolygon(x=x, cm=cm)
        index_list = list(range(p.ncaps))
        assert p.use_caps == 2**4 - 1
        use_caps = mng.set_use_caps(p, index_list, tol=1.0e-7)
        assert use_caps == 2**3 - 1
        x = np.array([[0.0, 0.0, 1.0],
                      [0.0, 1.0, 0.0],
                      [1.0, 0.0, 0.0],
                      [1.0-1.0e-8, 1.0e-8, 0.0]])
        cm = np.array([1.0, 1.0, 1.0, -1.0+1.0e-8])
        p = mng.ManglePolygon(x=x, cm=cm)
        index_list = list(range(p.ncaps))
        assert p.use_caps == 2**4 - 1
        use_caps = mng.set_use_caps(p, index_list, tol=1.0e-7)
        assert use_caps == 2**3 - 1
        use_caps = mng.set_use_caps(p, index_list, tol=1.0e-7,
                                    allow_neg_doubles=True)
        assert use_caps == 2**4 - 1

    def test_x_to_angles(self):
        a = mng.x_to_angles(np.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0],
                                      [1.0, 0.0, 0.0]]))
        assert np.allclose(a, np.array([[0.0, 0.0], [90.0, 90.0],
                                        [0.0, 90.0]]))
        a = mng.x_to_angles(np.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0],
                                      [1.0, 0.0, 0.0]]), latitude=True)
        assert np.allclose(a, np.array([[0.0, 90.0], [90.0, 0.0],
                                        [0.0, 0.0]]))

    def test_single_polygon(self):
        x = np.array([[0.0, 0.0, 1.0],
                      [0.0, 1.0, 0.0],
                      [1.0, 0.0, 0.0],
                      [1.0-1.0e-8, 1.0e-8, 0.0]])
        cm = np.array([1.0, 1.0, 1.0, -1.0+1.0e-8])
        p = mng.ManglePolygon(x=x, cm=cm)
        p2 = mng._single_polygon(p)
        assert id(p2) == id(p)
        pl = mng.PolygonList()
        pl.append(p)
        p2 = mng._single_polygon(pl)
        assert id(p2) == id(p)
        with pytest.raises(ValueError):
            pl.append(p2)
            _ = mng._single_polygon(pl)
        p = mng.read_fits_polygons(self.no_id_fits)
        p2 = mng._single_polygon(p)
        assert p2.ncaps == 4
        p2 = mng._single_polygon(p[0])
        assert p2.use_caps == 15


def fits_polygon_file():
    """Create a small test version of a FITS polygon file.
    """
    from datetime import date
    from sys import argv
    from astropy.io import fits
    from pydl import __version__ as pydlversion
    with fits.open(argv[1], uint=True) as hdulist:
        header0 = hdulist[0].header
        data = hdulist[1].data
    if 'DATE' in header0:
        header0['DATE'] = date.today().strftime('%Y-%m-%d')
    if 'IDLUTILS' in header0:
        header0['IDLUTILS'] = 'pydl-'+pydlversion
    hdu0 = fits.PrimaryHDU(header=header0)
    hdu1 = fits.BinTableHDU(data[0:20])
    hdulist2 = fits.HDUList([hdu0, hdu1])
    hdulist2.writeto('polygon.fits')
    return 0


if __name__ == '__main__':
    from sys import exit
    exit(fits_polygon_file())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the goddard/math directory in idlutils.
"""
import numpy as np
from scipy.special import legendre


def flegendre(x, m):
    """Compute the first `m` Legendre polynomials.

    Parameters
    ----------
    x : array-like
        Compute the Legendre polynomials at these abscissa values.
    m : :class:`int`
        The number of Legendre polynomials to compute.  For example, if
        :math:`m = 3`, :math:`P_0 (x)`, :math:`P_1 (x)` and :math:`P_2 (x)`
        will be computed.

    Returns
    -------
    :class:`numpy.ndarray`
        The values of the Legendre functions.
    """
    if isinstance(x, np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 1:
        raise ValueError('Number of Legendre polynomials must be at least 1.')
    try:
        dt = x.dtype
    except AttributeError:
        dt = np.float64
    leg = np.ones((m, n), dtype=dt)
    if m >= 2:
        leg[1, :] = x
    if m >= 3:
        for k in range(2, m):
            leg[k, :] = np.polyval(legendre(k), x)
    return leg



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.coord.
"""
from astropy import units as u
from astropy.coordinates import ICRS
from ..coord import SDSSMuNu, current_mjd, stripe_to_eta, stripe_to_incl


def test_current_mjd():
    assert current_mjd() > 50000.0


def test_munu_to_radec():
    munu = SDSSMuNu(mu=0.0*u.deg, nu=0.0*u.deg, stripe=10)
    radec = munu.transform_to(ICRS())
    assert radec.ra.value == 0.0
    assert radec.dec.value == 0.0


def test_radec_to_munu():
    radec = ICRS(ra=0.0*u.deg, dec=0.0*u.deg)
    munu = radec.transform_to(SDSSMuNu(stripe=10))
    assert munu.mu.value == 0.0
    assert munu.nu.value == 0.0
    assert munu.incl.value == 0.0


def test_stripe_to_eta():
    eta = stripe_to_eta(82)
    assert eta == -32.5


def test_stripe_to_incl():
    incl = stripe_to_incl(82)
    assert incl == 0.0
    incl = stripe_to_incl(10)
    assert incl == 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-


def uniq(x, index=None):
    """Replicates the IDL ``UNIQ()`` function.

    Returns the *subscripts* of the unique elements of an array.  The elements
    must actually be *sorted* before being passed to this function.  This can
    be done by sorting `x` explicitly or by passing the array subscripts that
    sort `x` as a second parameter.

    Parameters
    ----------
    x : array-like
        Search this array for unique items.
    index : array-like, optional
        This array provides the array subscripts that sort `x`.

    Returns
    -------
    array-like
        The subscripts of `x` that are the unique elements of `x`.

    Notes
    -----
    Given a sorted array, and assuming that there is a set of
    adjacent identical items, ``uniq()`` will return the subscript of the
    *last* unique item.  This charming feature is retained for
    reproducibility.

    References
    ----------
    https://www.nv5geospatialsoftware.com/docs/uniq.html

    Examples
    --------
    >>> import numpy as np
    >>> from pydl import uniq
    >>> data = np.array([ 1, 2, 3, 1, 5, 6, 1, 7, 3, 2, 5, 9, 11, 1 ])
    >>> print(uniq(np.sort(data)))
    [ 3  5  7  9 10 11 12 13]
    """
    from numpy import array, roll
    if index is None:
        indicies = (x != roll(x, -1)).nonzero()[0]
        if indicies.size > 0:
            return indicies
        else:
            return array([x.size - 1, ])
    else:
        q = x[index]
        indicies = (q != roll(q, -1)).nonzero()[0]
        if indicies.size > 0:
            return index[indicies]
        else:
            return array([q.size - 1, ], dtype=index.dtype)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-


def rebin(x, d, sample=False):
    """Resize `x` to new dimensions given by `d`.  The new dimensions must
    be integer multiples or factors of the original dimensions.

    Although there are some elegant solutions out there for rebinning, this
    function is intended to replace the IDL ``REBIN()`` function, which
    has a number of special properties:

    * It refuses to perform extrapolation when rebinning to a larger size
      in a particular dimension.
    * It can simultaneously rebin to a larger size in one dimension while
      rebinning to a smaller size in another dimension.

    Parameters
    ----------
    x : :class:`~numpy.ndarray`
        The array to resample.
    d : :class:`tuple`
        The new shape of the array.
    sample : :class:`bool`, optional
        If ``True``, nearest-neighbor techniques will be used instead of
        interpolation.

    Returns
    -------
    :class:`~numpy.ndarray`
        The resampled array.

    Raises
    ------
    :exc:`ValueError`
        If the new dimensions are incompatible with the algorithm.

    Warnings
    --------
    This function may not be 100% compatible with the IDL version
    *for integer inputs*. It is not possible at present to examine the details
    of the IDL code to determine the exact type manipulation that are used.
    For further discussion see Issue `#60`_.

    .. _`#60`: https://github.com/weaverba137/pydl/issues/60

    References
    ----------
    https://www.nv5geospatialsoftware.com/docs/rebin.html

    Examples
    --------
    >>> from numpy import arange, float32
    >>> from pydl import rebin
    >>> rebin(arange(10, dtype=float32), (5,)) # doctest: +NORMALIZE_WHITESPACE
    array([0.5, 2.5, 4.5, 6.5, 8.5], dtype=float32)
    >>> rebin(arange(5, dtype=float32), (10,)) # doctest: +NORMALIZE_WHITESPACE
    array([0. , 0.5, 1. , 1.5, 2. , 2.5, 3. , 3.5, 4. , 4. ], dtype=float32)
    """
    from numpy import floor, zeros
    d0 = x.shape
    if len(d0) != len(d):
        raise ValueError(("The new shape is incompatible with the " +
                          "original array.!"))
    for k in range(len(d0)):
        if d[k] > d0[k]:
            if d[k] % d0[k] != 0:
                raise ValueError(("{0:d} is not a multiple " +
                                  "of {1:d}!").format(d[k], d0[k]))
        elif d[k] == d0[k]:
            pass
        else:
            if d0[k] % d[k] != 0:
                raise ValueError(("{0:d} is not a multiple " +
                                  "of {1:d}!").format(d0[k], d[k]))
    xx = x.copy()
    new_shape = list(d0)
    for k in range(len(d0)):
        new_shape[k] = d[k]
        r = zeros(new_shape, dtype=xx.dtype)
        sliceobj0 = [slice(None)]*len(d0)
        sliceobj1 = [slice(None)]*len(d0)
        sliceobj = [slice(None)]*len(d)
        if d[k] > d0[k]:
            f = d0[k]/d[k]
            for i in range(d[k]):
                p = f*i
                fp = int(floor(p))
                sliceobj0[k] = slice(fp, fp + 1)
                sliceobj[k] = slice(i, i + 1)
                if sample:
                    r[tuple(sliceobj)] = xx[tuple(sliceobj0)]
                else:
                    if p < d0[k] - 1:
                        sliceobj1[k] = slice(fp + 1, fp + 2)
                        rshape = r[tuple(sliceobj)].shape
                        r[tuple(sliceobj)] = (xx[tuple(sliceobj0)].reshape(rshape) +
                                              (p - fp)*(xx[tuple(sliceobj1)] -
                                                        xx[tuple(sliceobj0)]).reshape(rshape))
                    else:
                        r[tuple(sliceobj)] = xx[tuple(sliceobj0)]
        elif d[k] == d0[k]:
            for i in range(d[k]):
                sliceobj0[k] = slice(i, i + 1)
                sliceobj[k] = slice(i, i + 1)
                r[tuple(sliceobj)] = xx[tuple(sliceobj0)]
        else:
            f = d0[k]//d[k]
            for i in range(d[k]):
                sliceobj[k] = slice(i, i + 1)
                if sample:
                    fp = int(floor(f*i))
                    sliceobj0[k] = slice(fp, fp + 1)
                    r[tuple(sliceobj)] = xx[tuple(sliceobj0)]
                else:
                    sliceobj0[k] = slice(int(f*i), int(f*(i+1)))
                    rshape = r[tuple(sliceobj)].shape
                    rr = xx[tuple(sliceobj0)].sum(k).reshape(rshape)
                    if xx.dtype.kind == 'u' or xx.dtype.kind == 'i':
                        r[tuple(sliceobj)] = rr//f
                    else:
                        r[tuple(sliceobj)] = rr/f
        xx = r
    return r



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.photoop.photoobj.
"""
import pytest
import numpy as np
from ..photoobj import sdss_calibv, unwrap_objid


def test_sdss_calibv():
    assert np.allclose(0.2650306748466258, sdss_calibv().value)


def test_unwrap_objid():
    objid = unwrap_objid(np.array([1237661382772195474]))
    assert objid.skyversion == 2
    assert objid.rerun == 301
    assert objid.run == 3704
    assert objid.camcol == 3
    assert objid.firstfield == 0
    assert objid.frame == 91
    assert objid.id == 146
    objid = unwrap_objid(np.array(['1237661382772195474']))
    assert objid.skyversion == 2
    assert objid.rerun == 301
    assert objid.run == 3704
    assert objid.camcol == 3
    assert objid.firstfield == 0
    assert objid.frame == 91
    assert objid.id == 146
    objid = unwrap_objid(np.array([587722984180548043]))
    assert objid.skyversion == 1
    assert objid.rerun == 40
    assert objid.run == 752
    assert objid.camcol == 5
    assert objid.firstfield == 1
    assert objid.frame == 618
    assert objid.id == 459
    with pytest.raises(ValueError):
        objid = unwrap_objid(np.array([3.14159]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the spec2d directory in idlspec2d.
"""
from warnings import warn
import numpy as np
from scipy.special import erf
from astropy import log
from astropy.io import ascii
from astropy.utils.data import get_pkg_data_filename
from . import Pydlspec2dException, Pydlspec2dUserWarning
from .. import smooth
from ..pydlutils.bspline import iterfit
from ..pydlutils.image import djs_maskinterp
from ..pydlutils.math import djs_median
from ..pydlutils.sdss import sdss_flagval
from ..pydlutils.trace import traceset2xy, xy2traceset
from ..goddard.astro import vactoair


def aesthetics(flux, invvar, method='traditional'):
    """Add nice values to a spectrum where it is masked.

    Parameters
    ----------
    flux : :class:`numpy.ndarray`
        The spectrum to clean up.
    invvar : :class:`numpy.ndarray`
        Inverse variance of the spectrum.
    method : { 'traditional', 'noconst', 'mean', 'damp', 'nothing' }, optional
        Apply this method to clean up the spectrum.  Default is 'traditional'.

    Returns
    -------
    :class:`numpy.ndarray`
        A cleaned-up spectrum.
    """
    badpts = invvar == 0
    if badpts.any():
        if method == 'traditional':
            newflux = djs_maskinterp(flux, invvar == 0, const=True)
        elif method == 'noconst':
            newflux = djs_maskinterp(flux, invvar == 0)
        elif method == 'mean':
            newflux = flux.copy()
            goodpts = invvar > 0
            newflux[~goodpts] = newflux[goodpts].mean()
        elif method == 'damp':
            l = 250  # damping length in pixels
            goodpts = invvar.nonzero()[0]
            nflux = flux.size
            mingood = goodpts.min()
            maxgood = goodpts.max()
            newflux = djs_maskinterp(flux, invvar == 0, const=True)
            pixels = np.arange(nflux, dtype='f')
            if mingood > 0:
                damp1 = float(min(mingood, l))
                newflux *= 0.5*(1.0+erf((pixels-mingood)/damp1))
            if maxgood < (nflux - 1):
                damp2 = float(min(maxgood, l))
                newflux *= 0.5*(1.0+erf((maxgood-pixels)/damp2))
        elif method == 'nothing':
            newflux = flux.copy()
        else:
            raise Pydlspec2dException("Unknown method: {0}".format(method))
        return newflux
    else:
        return flux


def combine1fiber(inloglam, objflux, newloglam, objivar=None, verbose=False,
                  **kwargs):
    """Combine several spectra of the same object, or resample a single spectrum.

    Parameters
    ----------
    inloglam : :class:`numpy.ndarray`
        Vector of log wavelength.
    objflux : :class:`numpy.ndarray`
        Input flux.
    newloglam : :class:`numpy.ndarray`
        Output wavelength pixels, vector of log wavelength.
    objivar : :class:`numpy.ndarray`, optional
        Inverse variance of the flux.
    verbose : :class:`bool`, optional
        If ``True``, set log level to DEBUG.

    Returns
    -------
    :class:`tuple` of :class:`numpy.ndarray`
        The resulting flux and inverse variance.

    Raises
    ------
    :exc:`ValueError`
        If input dimensions don't match.
    """
    #
    # Log
    #
    # log.enable_warnings_logging()
    if verbose:
        log.setLevel('DEBUG')
    #
    # Check that dimensions of inputs are valid.
    #
    npix = inloglam.size
    nfinalpix = len(newloglam)
    if objflux.shape != inloglam.shape:
        raise ValueError('Dimensions of inloglam and objflux do not agree.')
    if objivar is not None:
        if objivar.shape != inloglam.shape:
            raise ValueError('Dimensions of inloglam and objivar do not agree.')
    if 'finalmask' in kwargs:
        if kwargs['finalmask'].shape != inloglam.shape:
            raise ValueError('Dimensions of inloglam and finalmask do not agree.')
    if 'indisp' in kwargs:
        if kwargs['indisp'].shape != inloglam.shape:
            raise ValueError('Dimensions of inloglam and indisp do not agree.')
    #
    # Set defaults
    #
    EPS = np.finfo(np.float32).eps
    if 'binsz' in kwargs:
        binsz = kwargs['binsz']
    else:
        if inloglam.ndim == 2:
            binsz = inloglam[0, 1] - inloglam[0, 0]
        else:
            binsz = inloglam[1] - inloglam[0]
    if 'nord' in kwargs:
        nord = kwargs['nord']
    else:
        nord = 3
    if 'bkptbin' in kwargs:
        bkptbin = kwargs['bkptbin']
    else:
        bkptbin = 1.2 * binsz
    if 'maxsep' in kwargs:
        maxsep = kwargs['maxsep']
    else:
        maxsep = 2.0 * binsz
    if inloglam.ndim == 1:
        #
        # Set specnum = 0 for all elements
        #
        nspec = 1
        specnum = np.zeros(inloglam.shape, dtype=inloglam.dtype)
    else:
        nspec, ncol = inloglam.shape
        specnum = np.tile(np.arange(nspec), ncol).reshape(ncol, nspec).transpose()
    #
    # Use fullcombmask for modifying the pixel masks in the original input files.
    #
    fullcombmask = np.zeros(npix)
    newflux = np.zeros(nfinalpix, dtype=inloglam.dtype)
    newmask = np.zeros(nfinalpix, dtype='i4')
    newivar = np.zeros(nfinalpix, dtype=inloglam.dtype)
    newdisp = np.zeros(nfinalpix, dtype=inloglam.dtype)
    newsky = np.zeros(nfinalpix, dtype=inloglam.dtype)
    newdispweight = np.zeros(nfinalpix, dtype=inloglam.dtype)
    if objivar is None:
        nonzero = np.arange(npix, dtype='i4')
        ngood = npix
    else:
        nonzero = (objivar.ravel() > 0).nonzero()[0]
        ngood = nonzero.size
    #
    # ormask is needed to create andmask
    #
    andmask = np.zeros(nfinalpix, dtype='i4')
    ormask = np.zeros(nfinalpix, dtype='i4')
    if ngood == 0:
        #
        # In this case of no good points, set the nodata bit everywhere.
        # Also if noplug is set in the first input bit-mask, assume it
        # should be set everywhere in the output bit masks.  No other bits
        # are set.
        #
        warn('No good points!', Pydlspec2dUserWarning)
        bitval = sdss_flagval('SPPIXMASK', 'NODATA')
        if 'finalmask' in kwargs:
            bitval |= (sdss_flagval('SPPIXMASK', 'NOPLUG') *
                       (kwargs['finalmask'][0] & sdss_flagval('SPPIXMASK', 'NODATA')))
        andmask = andmask | bitval
        ormask = ormask | bitval
        return (newflux, newivar)
    else:
        #
        # Now let's break sorted wavelengths into groups where pixel
        # separations are larger than maxsep.
        #
        inloglam_r = inloglam.ravel()
        isort = nonzero[inloglam_r[nonzero].argsort()]
        wavesort = inloglam_r[isort]
        padwave = np.insert(wavesort, 0, wavesort.min() - 2.0*maxsep)
        padwave = np.append(padwave, wavesort.max() + 2.0*maxsep)
        ig1 = ((padwave[1:ngood+1]-padwave[0:ngood]) > maxsep).nonzero()[0]
        ig2 = ((padwave[2:ngood+2]-padwave[1:ngood+1]) > maxsep).nonzero()[0]
        if ig1.size != ig2.size:
            raise ValueError('Grouping tricks did not work!')
        #
        # Avoid flux-dependent bias when combining multiple spectra.
        # This call to djs_median contains a width that is both floating-point
        # and even, which is very strange.
        #
        if objivar is not None and objivar.ndim > 1:
            saved_objivar = objivar
            for spec in range(nspec):
                igood = (objivar[spec, :] > 0).nonzero()[0]
                if igood.size > 0:
                    # objivar[spec, igood] = djs_median(saved_objivar[spec, igood], width=100.)
                    objivar[spec, igood] = djs_median(saved_objivar[spec, igood], width=101)
        else:
            saved_objivar = None
        for igrp in range(ig1.size):
            ss = isort[ig1[igrp]:ig2[igrp]+1]
            if ss.size > 2:
                if objivar is None:
                    #
                    # Fit without variance
                    #
                    sset, bmask = iterfit(inloglam_r[ss],
                                          objflux.ravel()[ss],
                                          nord=nord, groupbadpix=True,
                                          requiren=1, bkspace=bkptbin,
                                          silent=True)
                else:
                    #
                    # Fit with variance
                    #
                    sset, bmask = iterfit(inloglam_r[ss],
                                          objflux.ravel()[ss],
                                          invvar=objivar.ravel()[ss],
                                          nord=nord, groupbadpix=True,
                                          requiren=1, bkspace=bkptbin,
                                          silent=True)
                if np.sum(np.absolute(sset.coeff)) == 0:
                    sset = None
                    bmask = np.zeros(len(ss))
                    warn('All B-spline coefficients have been set to zero!',
                         Pydlspec2dUserWarning)
            else:
                bmask = np.zeros(len(ss))
                sset = None
                warn('Not enough data for B-spline fit!', Pydlspec2dUserWarning)
            inside = ((newloglam >= (inloglam_r[ss]).min()-EPS) &
                      (newloglam <= (inloglam_r[ss]).max()+EPS)).nonzero()[0]
            #
            # It is possible for numinside to be zero, if the input data points
            # span an extremely small wavelength range, within which there are
            # no output wavelengths.
            #
            if sset is not None and len(inside) > 0:
                newflux[inside], bvalumask = sset.value(newloglam[inside])
                if bvalumask.any():
                    newmask[inside[bvalumask]] = 1
                log.debug('Masked {0:d} of {1:d} pixels.'.format((1-bmask).sum(), bmask.size))
                #
                # Determine which pixels should be masked based upon the spline
                # fit. Set the combinerej bit.
                #
                ireplace = ~bmask
                if ireplace.any():
                    #
                    # The following would replace the original flux values of
                    # masked pixels with b-spline evaluations.
                    #
                    # objflux[ss[ireplace]] = sset.value(inloglam[ss[ireplace]])
                    #
                    # Set the inverse variance of these pixels to zero.
                    #
                    if objivar is not None:
                        objivar.ravel()[ss[ireplace]] = 0.0
                        log.debug('Replaced {0:d} pixels in objivar.'.format(len(ss[ireplace])))
                    if 'finalmask' in kwargs:
                        kwargs['finalmask'][ss[ireplace]] = (kwargs['finalmask'][ss[ireplace]] |
                                                             sdss_flagval('SPPIXMASK', 'COMBINEREJ'))
            fullcombmask[ss] = bmask
        #
        # Restore objivar
        #
        if saved_objivar is not None:
            objivar = saved_objivar * (objivar > 0)
        #
        # Combine inverse variance and pixel masks.
        #
        # Start with all bits set in andmask
        #
        andmask[:] = -1
        for j in range(int(specnum.max())+1):
            these = (specnum.ravel() == j).nonzero()[0]
            if these.any():
                inbetween = ((newloglam >= inloglam_r[these].min()) &
                             (newloglam <= inloglam_r[these].max()))
                if inbetween.any():
                    jnbetween = inbetween.nonzero()[0]
                    #
                    # Conserve inverse variance by doing a linear interpolation
                    # on that quantity.
                    #
                    result = np.interp(newloglam[jnbetween], inloglam_r[these],
                                       (objivar.ravel()[these] *
                                        fullcombmask[these]))
                    #
                    # Grow the fullcombmask below to reject any new sampling
                    # containing even a partial masked pixel.
                    #
                    smask = np.interp(newloglam[jnbetween], inloglam_r[these],
                                      fullcombmask[these].astype(inloglam.dtype))
                    result *= smask >= (1.0 - EPS)
                    newivar[jnbetween] += result*newmask[jnbetween]
                lowside = np.floor((inloglam_r[these]-newloglam[0])/binsz).astype('i4')
                highside = lowside + 1
                if 'finalmask' in kwargs:
                    andmask[lowside] &= kwargs['finalmask'][these]
                    andmask[highside] &= kwargs['finalmask'][these]
                    ormask[lowside] |= kwargs['finalmask'][these]
                    ormask[highside] |= kwargs['finalmask'][these]
                #
                # Combine the dispersions + skies in the dumbest way possible
                # [sic].
                #
                if 'indisp' in kwargs:
                    newdispweight[jnbetween] += result
                    newdisp[jnbetween] += (result *
                                           np.interp(newloglam[jnbetween],
                                                     inloglam_r[these],
                                                     kwargs['indisp'].ravel()[these]))
                    newsky[jnbetween] += (result *
                                          np.interp(newloglam[jnbetween],
                                                    inloglam_r[these],
                                                    kwargs['skyflux'].ravel()[these]))
        if 'indisp' in kwargs:
            newdisp /= newdispweight + (newdispweight == 0)
            newsky /= newdispweight + (newdispweight == 0)
    #
    # Grow regions where 3 or more pixels are rejected together ???
    #
    foo = smooth(newivar, 3)
    badregion = np.absolute(foo) < EPS
    # badregion = foo == 0.0
    if badregion.any():
        warn('Growing bad pixel region, {0:d} pixels found.'.format(badregion.sum()),
             Pydlspec2dUserWarning)
        ibad = badregion.nonzero()[0]
        lowerregion = np.where(ibad-2 < 0, 0, ibad-2)
        upperregion = np.where(ibad+2 > nfinalpix-1, nfinalpix-1, ibad+2)
        newivar[lowerregion] = 0.0
        newivar[upperregion] = 0.0
    #
    # Replace NaNs in combined spectra; this should really never happen.
    #
    inff = ((~np.isfinite(newflux)) | (~np.isfinite(newivar)))
    if inff.any():
        warn('{0:d} NaNs in combined spectra.'.format(inff.sum()),
             Pydlspec2dUserWarning)
        newflux[inff] = 0.0
        newivar[inff] = 0.0
    #
    # Interpolate over masked pixels, just for aesthetic purposoes.
    #
    goodpts = newivar > 0
    if 'aesthetics' in kwargs:
        amethod = kwargs['aesthetics']
    else:
        amethod = 'traditional'
    newflux = aesthetics(newflux, newivar, method=amethod)
    # if 'interpolate' in kwargs:
    #     newflux = pydlutils.image.djs_maskinterp(newflux,~goodpts,const=True)
    # else:
    #     newflux[~goodpts] = newflux[goodpts].mean()
    if goodpts.any():
        minglam = newloglam[goodpts].min()
        maxglam = newloglam[goodpts].max()
        ibad = ((newloglam < minglam) | (newloglam > maxglam))
        if ibad.any():
            ormask[ibad] |= sdss_flagval('SPPIXMASK', 'NODATA')
            andmask[ibad] |= sdss_flagval('SPPIXMASK', 'NODATA')
    #
    # Replace values of -1 in the andmask with 0.
    #
    andmask *= (andmask != -1)
    return (newflux, newivar)


def filter_thru(flux, waveimg=None, wset=None, mask=None,
                filter_prefix='sdss_jun2001', toair=False):
    """Compute throughput in SDSS filters.

    Parameters
    ----------
    flux : array-like
        Spectral flux.
    waveimg : array-like, optional
        Full wavelength solution, with the same shape as `flux`.
    wset : :class:`~pydl.pydlutils.trace.TraceSet`, optional
        A trace set containing the wavelength solution.  Must be specified
        if `waveimg` is not specified.
    mask : array-like, optional
        Interpolate over pixels where `mask` is non-zero.
    filter_prefix : :class:`str`, optional
        Specifies a set of filter curves.
    toair : :class:`bool`, optional
        If ``True``, convert the wavelengths to air from vacuum before computing.

    Returns
    -------
    array-like
        Integrated flux in the filter bands.

    Raises
    ------
    :exc:`ValueError`
        If neither `waveimg` nor `wset` are set.
    """
    nTrace, nx = flux.shape
    if filter_prefix != 'sdss_jun2001':
        raise ValueError("Filters other than {0} are not available!".format('sdss_jun2001'))
    ffiles = [_get_pkg_filename_compat('data/filters/{0}_{1}_atm.dat'.format(filter_prefix, f),
                                       package='pydl.pydlutils') for f in 'ugriz']
    if waveimg is None and wset is None:
        raise ValueError("Either waveimg or wset must be specified!")
    if waveimg is None:
        pixnorm, logwave = traceset2xy(wset)
        waveimg = 10**logwave
    if toair:
        newwaveimg = vactoair(waveimg)
    else:
        newwaveimg = waveimg
    logwave = np.log10(newwaveimg)
    diffx = np.outer(np.ones((nTrace,), dtype=flux.dtype), np.arange(nx-1, dtype=flux.dtype))
    diffy = logwave[:, 1:] - logwave[:, 0:nx-1]
    diffset = xy2traceset(diffx, diffy, ncoeff=4, xmin=0, xmax=nx-1)
    pixnorm, logdiff = traceset2xy(diffset)
    logdiff = np.absolute(logdiff)
    if mask is not None:
        flux_interp = djs_maskinterp(flux, mask, axis=0)
    res = np.zeros((nTrace, len(ffiles)), dtype=flux.dtype)
    for i, f in enumerate(ffiles):
        filter_data = ascii.read(f, comment='#.*', names=('lam', 'respt',
                                 'resbig', 'resnoa', 'xatm'))
        filtimg = logdiff * np.interp(newwaveimg.flatten(),
                                      filter_data['lam'].data,
                                      filter_data['respt'].data).reshape(logdiff.shape)
        if mask is not None:
            res[:, i] = (flux_interp * filtimg).sum(1)
        else:
            res[:, i] = (flux * filtimg).sum(1)
        sumfilt = filtimg.sum(1)
        res[:, i] = res[:, i] / (sumfilt + (sumfilt <= 0).astype(sumfilt.dtype))
    return res


def _get_pkg_filename_compat(filename, package):
    """Astropy 1.0.x/LTS does not accept the 'package' argument.
    """
    try:
        f = get_pkg_data_filename(filename, package=package)
    except TypeError:
        from pkg_resources import resource_filename
        f = resource_filename(package, filename)
    return f



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the sdssio directory of photoop.
"""
import os
import numpy as np


#
# Filename formats used by sdss_name and sdss_path
#
_name_formats = {
    'apObj': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'calibMatch': "{ftype}-{run:06d}-{camcol:1d}.fits",
    'calibPhotom': "{ftype}-{run:06d}-{camcol:1d}.fits",
    'calibPhotomGlobal': "{ftype}-{run:06d}-{camcol:1d}.fits",
    'fakeIdR': "idR-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'fpAtlas': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fit",
    'fpBIN': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'fpC': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'fpFieldStat': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fit",
    'fpM': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'fpObjc': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fit",
    'hoggObj': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fits",
    'idFF': "{ftype}-{run:06d}-{filter}{camcol:1d}.fit",
    'idR': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'idRR': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'psBB': "{ftype}-{run:06d}-{filter}{camcol:1d}-{field:04d}.fit",
    'psFF': "{ftype}-{run:06d}-{filter}{camcol:1d}.fit",
    'psField': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fit",
    'reObjGlobal': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fits",
    'reObjRun': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fits",
    'reObjTmp': "{ftype}-{run:06d}-{camcol:1d}-{field:04d}.fits",
    'tsField': "{ftype}-{run:06d}-{camcol:1d}-{rerun}-{field:04d}.fit"}


_path_formats = {
    'apObj': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'calibMatch': "{redux}/{rerun}/{run:d}/nfcalib",
    'calibPhotom': "{redux}/{rerun}/{run:d}/nfcalib",
    'calibPhotomGlobal': "{calib}/{rerun}/{run:d}/nfcalib",
    'fakeIdR': "{data}/{run:d}/fake_fields/{camcol:1d}",
    'fpAtlas': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpBIN': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpC': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpFieldStat': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpM': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpObjc': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'hoggObj': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'idFF': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'idR': "{data}/{run:d}/fields/{camcol:1d}",
    'idRR': "{data}/{run:d}/fields/{camcol:1d}",
    'psBB': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'psFF': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'psField': "{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'reObjGlobal': "{resolve}/{rerun}/{run:d}/resolve/{camcol:1d}",
    'reObjRun': "{redux}/{rerun}/{run:d}/resolve/{camcol:1d}",
    'reObjTmp': "{resolve}/{rerun}/{run:d}/resolve/{camcol:1d}",
    'tsField': "{redux}/{rerun}/{run:d}/calibChunks/{camcol:1d}"}


def filtername(f):
    """Return the name of a filter given its number.

    Parameters
    ----------
    f : :class:`int`
        The filter number.

    Returns
    -------
    :class:`str`
        The corresponding filter name.

    Examples
    --------
    >>> filtername(0)
    'u'
    """
    if isinstance(f, (str,)):
        return f
    fname = ('u', 'g', 'r', 'i', 'z')
    return fname[f]


def filternum(filt='foo'):
    """Return index number for SDSS filters either from a number or name.

    Parameters
    ----------
    filt : :class:`str`
        The filter name.

    Returns
    -------
    :class:`int`
        The corresponding filter number

    Raises
    ------
    :exc:`KeyError`
        If `filt` is not a valid filter name.

    Examples
    --------
    >>> filternum('g')
    1
    """
    if filt == 'foo':
        return list(range(5))
    else:
        filters = {'u': 0, 'g': 1, 'r': 2, 'i': 3, 'z': 4}
        return filters[filt]


def sdss_calib(run, camcol, field, rerun='', **kwargs):
    """Read photometric calibration solutions from calibPhotom or
    calibPhotomGlobal files.

    Parameters
    ----------
    run : :class:`int`
        Photo run number
    camcol : :class:`int`
        Camcol number
    field : :class:`int`
        Field number
    rerun : :class:`str`, optional
        Photometric reduction number, as a string.

    Returns
    -------
    :class:`dict`
        A dictionary containing the 'NMGYPERCOUNT' keyword.

    Notes
    -----
    Currently, this is just a placeholder.
    """
    return {'NMGYPERCOUNT': 1.0}


def sdss_name(ftype, run, camcol, field, rerun='', thisfilter='r',
              no_path=False):
    """Return the name of an SDSS data file including path.

    Parameters
    ----------
    ftype : :class:`str`
        The general type of the file, for example ``'reObj'``
    run : :class:`int`
        The run number.
    camcol : :class:`int`
        The camcol number.
    field : :class:`int`
        The field number
    rerun : :class:`str`, optional
        If necessary, set the rerun name using this argument.
    thisfilter : :class:`int` or :class:`str`, optional
        If necessary, set the filter using this argument.
    no_path : :class:`bool`, optional
        Normally, sdss_name returns the full path.  If `no_path` is ``True``,
        only the basename of the file is returned.

    Returns
    -------
    :class:`str`
        The full file name, normally including the full path.

    Raises
    ------
    :exc:`KeyError`
        If the file type is unknown.
    """
    if ftype == 'reObj':
        if 'PHOTO_RESOLVE' in os.environ:
            myftype = 'reObjGlobal'
        else:
            myftype = 'reObjRun'
    else:
        myftype = ftype
    thisfilter = filtername(thisfilter)
    indict = {'ftype': myftype, 'run': run, 'camcol': camcol, 'field': field,
              'filter': thisfilter, 'rerun': rerun}
    try:
        fullname = _name_formats[myftype].format(**indict)
    except KeyError:
        raise KeyError("Unknown FTYPE = {0}".format(myftype))
    if not no_path:
        datadir = sdss_path(myftype, run, camcol, rerun)
        fullname = os.path.join(datadir, fullname)
    return fullname


def sdss_path(ftype, run, camcol=0, rerun=''):
    """Return the path name for SDSS data assuming SAS directory structure.

    Parameters
    ----------
    ftype : :class:`str`
        The general type of the file, for example ``'reObj'``
    run : :class:`int`
        The run number.
    camcol : :class:`int`, optional
        If necessary, set the camcol number using this argument.
    rerun : :class:`str`, optional
        If necessary, set the rerun name using this argument.

    Returns
    -------
    :class:`str`
        The directory in which file `ftype` lives.

    Raises
    ------
    :exc:`KeyError`
        If the file type is unknown.

    """
    indict = {'run': run,
              'camcol': camcol,
              'rerun': rerun,
              'calib': os.getenv('PHOTO_CALIB'),
              'data': os.getenv('PHOTO_DATA'),
              'photoobj': os.getenv('BOSS_PHOTOOBJ'),
              'redux': os.getenv('PHOTO_REDUX'),
              'resolve': os.getenv('PHOTO_RESOLVE'),
              'sky': os.getenv('PHOTO_SKY'),
              'sweep': os.getenv('PHOTO_SWEEP')}
    try:
        datadir = _path_formats[ftype].format(**indict)
    except KeyError:
        raise KeyError("Unknown FTYPE = {0}".format(ftype))
    return datadir


def sdssflux2ab(flux, magnitude=False, ivar=False):
    """Convert the SDSS calibrated fluxes (magnitudes) into AB fluxes
    (magnitudes).

    Parameters
    ----------
    flux : :class:`numpy.ndarray`
        Array of calibrated fluxes or SDSS magnitudes with 5 columns,
        corresponding to the 5 filters *u*, *g*, *r*, *i*, *z*.
    magnitude : :class:`bool`, optional
        If set to ``True``, then assume `flux` are SDSS magnitudes instead of
        linear flux units.
    ivar : :class:`numpy.ndarray`, optional
        If set, the input fluxes are actually inverse variances.

    Returns
    -------
    :class:`numpy.ndarray`
        Array of fluxes or magnitudes on the AB system.

    Notes
    -----
    Uses the conversions posted by D.Hogg (sdss-calib/845)::

        u(AB,2.5m) = u(2.5m) - 0.042
        g(AB,2.5m) = g(2.5m) + 0.036
        r(AB,2.5m) = r(2.5m) + 0.015
        i(AB,2.5m) = i(2.5m) + 0.013
        z(AB,2.5m) = z(2.5m) - 0.002
    """
    #
    # Correction vector, adjust this as necessary
    #
    correction = np.array([-0.042, 0.036, 0.015, 0.013, -0.002])
    rows, cols = flux.shape
    abflux = flux.copy()
    if magnitude:
        for i in range(rows):
            abflux[i, :] += correction
    else:
        factor = 10.0**(-correction/2.5)
        if ivar:
            factor = 1.0/factor**2
        for i in range(rows):
            abflux[i, :] *= factor
    return abflux



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import pytest
import numpy as np
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from ..trace import (fchebyshev, fchebyshev_split, fpoly, func_fit,
                     TraceSet, traceset2xy, xy2traceset)
from .. import PydlutilsException


class TestTrace(object):
    """Test the functions in pydl.pydlutils.trace.
    """

    def setup_method(self):
        # extracted from spFrame-b1-00057618.fits

        self.sdss = fits.open(get_pkg_data_filename('t/sdss_traceset.fits'))
        # extracted from spFrame-r1-00180406.fits
        self.boss = fits.open(get_pkg_data_filename('t/boss_traceset.fits'))
        return

    def teardown_method(self):
        self.sdss.close()
        self.boss.close()
        return

    def test_fchebyshev(self):
        x = np.array([-1, -0.5, 0, 0.5, 1], dtype='d')
        #
        # Test order
        #
        with pytest.raises(ValueError):
            f = fchebyshev(x, 0)
        #
        # m = 1
        #
        f = fchebyshev(x, 1)
        assert (f == np.ones((1, x.size), dtype='d')).all()
        #
        # m = 2
        #
        f = fchebyshev(x, 2)
        foo = np.ones((2, x.size), dtype='d')
        foo[1, :] = x
        assert np.allclose(f, foo)
        #
        # m = 3
        #
        f = fchebyshev(x, 3)
        foo = np.ones((3, x.size), dtype='d')
        foo[1, :] = x
        foo[2, :] = (2.0*x**2 - 1.0)
        assert np.allclose(f, foo)
        #
        # m = 4
        #
        f = fchebyshev(x, 4)
        foo = np.ones((4, x.size), dtype='d')
        foo[1, :] = x
        foo[2, :] = (2.0*x**2 - 1.0)
        foo[3, :] = (4.0*x**3 - 3.0*x)
        assert np.allclose(f, foo)
        #
        # random float
        #
        f = fchebyshev(2.88, 3)
        assert np.allclose(f, np.array([[1.00], [2.88], [15.5888]]))

    def test_fchebyshev_split(self):
        x = np.array([-1, -0.5, 0, 0.5, 1], dtype='d')
        #
        # Test order
        #
        with pytest.raises(ValueError):
            f = fchebyshev_split(x, 0)
        with pytest.raises(ValueError):
            f = fchebyshev_split(x, 1)
        #
        # m = 2
        #
        f = fchebyshev_split(x, 2)
        foo = np.ones((2, x.size), dtype='d')
        foo[0, :] = (x >= 0).astype(x.dtype)
        assert np.allclose(f, foo)
        #
        # m = 3
        #
        f = fchebyshev_split(x, 3)
        foo = np.ones((3, x.size), dtype='d')
        foo[0, :] = (x >= 0).astype(x.dtype)
        foo[2, :] = x
        assert np.allclose(f, foo)
        #
        # m = 4
        #
        f = fchebyshev_split(x, 4)
        foo = np.ones((4, x.size), dtype='d')
        foo[0, :] = (x >= 0).astype(x.dtype)
        foo[2, :] = x
        foo[3, :] = (2.0*x**2 - 1.0)
        assert np.allclose(f, foo)
        #
        # m = 5
        #
        f = fchebyshev_split(x, 5)
        foo = np.ones((5, x.size), dtype='d')
        foo[0, :] = (x >= 0).astype(x.dtype)
        foo[2, :] = x
        foo[3, :] = (2.0*x**2 - 1.0)
        foo[4, :] = (4.0*x**3 - 3.0*x)
        assert np.allclose(f, foo)
        #
        # random float
        #
        f = fchebyshev_split(2.88, 3)
        assert np.allclose(f, np.array([[1.00], [1.00], [2.88]]))

    def test_fpoly(self):
        x = np.array([-1, -0.5, 0, 0.5, 1], dtype='d')
        #
        # Test order
        #
        with pytest.raises(ValueError):
            _ = fpoly(x, 0)
        #
        # m = 1
        #
        f = fpoly(x, 1)
        assert (f == np.ones((1, x.size), dtype='d')).all()
        #
        # m = 2
        #
        f = fpoly(x, 2)
        foo = np.ones((2, x.size), dtype='d')
        foo[1, :] = x
        assert np.allclose(f, foo)
        #
        # m = 3
        #
        f = fpoly(x, 3)
        foo = np.ones((3, x.size), dtype='d')
        foo[1, :] = x
        foo[2, :] = x**2
        assert np.allclose(f, foo)
        #
        # m = 4
        #
        f = fpoly(x, 4)
        foo = np.ones((4, x.size), dtype='d')
        foo[1, :] = x
        foo[2, :] = x**2
        foo[3, :] = x**3
        assert np.allclose(f, foo)
        #
        # random float
        #
        f = fpoly(2.88, 3)
        assert np.allclose(f, np.array([[1.00], [2.88], [8.29440]]))

    def test_func_fit(self):
        np.random.seed(137)
        x = np.linspace(-5, 5, 50)
        y = x**2 + 2*x + 1 + 0.05*np.random.randn(50)
        #
        # Bad inputs
        #
        with pytest.raises(ValueError):
            _ = func_fit(x, np.array([1, 2, 3]), 3)
        with pytest.raises(ValueError):
            _ = func_fit(x, y, 3, invvar=np.array([0, 0, 0]))
        with pytest.raises(ValueError):
            _ = func_fit(x, y, 3, inputfunc=np.array([1.0, 1.0, 1.0]))
        with pytest.raises(KeyError):
            _ = func_fit(x, y, 3, function_name='npoly')
        #
        # No good points
        #
        invvar = np.zeros(x.shape, dtype=x.dtype)
        res, yfit = func_fit(x, y, 3, invvar=invvar)
        assert (res == np.zeros((3,), dtype=x.dtype)).all()
        assert (yfit == 0*x).all()
        #
        # One good point
        #
        invvar[2] = 1.0
        res, yfit = func_fit(x, y, 3, invvar=invvar)
        assert (invvar > 0).nonzero()[0][0] == 2
        assert res[0] == y[2]
        assert (yfit == y[2]).all()
        #
        # Various points
        #
        invvar = 1.0/(np.random.random(x.shape)**2)
        assert (invvar < 2).any()
        invvar[invvar < 2] = 0
        res, yfit = func_fit(x, y, 3, invvar=invvar, function_name='poly')
        # assert np.allclose(res,np.array([0.99665423, 1.9945388, 1.00172303]))
        assert np.allclose(res, np.array([0.99996197, 1.99340315, 1.00148004]))
        #
        # Fixed points
        #
        res, yfit = func_fit(x, y, 3, invvar=invvar, function_name='poly',
                             ia=np.array([False, True, True]),
                             inputans=np.array([1.0, 0, 0]))
        # assert np.allclose(res,np.array([1., 1.99454782, 1.00149949]))
        assert np.allclose(res, np.array([1., 1.99340359, 1.00147743]))
        res, yfit = func_fit(x, y, 3, invvar=invvar, function_name='poly',
                             ia=np.array([False, True, False]),
                             inputans=np.array([1.0, 0, 1.0]))
        # assert np.allclose(res,np.array([1., 1.99403239, 1.]))
        assert np.allclose(res, np.array([1., 1.99735654, 1.]))
        #
        # inputfunc
        #
        res, yfit = func_fit(x, y, 3, invvar=invvar, function_name='poly',
                             inputfunc=np.ones(x.shape, dtype=x.dtype))
        # assert np.allclose(res,np.array([0.99665423, 1.9945388, 1.00172303]))
        assert np.allclose(res, np.array([0.99996197, 1.99340315, 1.00148004]))
        #
        # Generate inputans
        #
        y = x**2 + 2*x + 0.05*np.random.randn(50)
        res, yfit = func_fit(x, y, 3, invvar=invvar, function_name='poly',
                             ia=np.array([False, True, True]))
        assert np.allclose(res, np.array([0., 1.99994188, 0.99915111]))

    def test_traceset_sdss(self):
        tset = TraceSet(self.sdss[1].data)
        assert tset.func == 'legendre'
        assert tset.coeff.shape == (320, 5)
        assert not tset.has_jump
        assert tset.xRange == tset.xmax - tset.xmin
        assert tset.nx == int(tset.xmax - tset.xmin + 1)
        assert tset.xmid == 0.5 * (tset.xmin + tset.xmax)
        x, y = traceset2xy(tset)
        tset2 = xy2traceset(x, y, ncoeff=tset.ncoeff)
        assert tset2.xmin == tset.xmin
        assert tset2.xmax == tset.xmax
        assert tset2.func == tset.func
        assert np.allclose(tset2.coeff, tset.coeff)

    def test_traceset_boss(self):
        tset = TraceSet(self.boss[1].data)
        assert tset.func == 'legendre'
        assert tset.coeff.shape == (500, 6)
        assert tset.has_jump
        assert tset.xRange == tset.xmax - tset.xmin
        assert tset.nx == int(tset.xmax - tset.xmin + 1)
        assert tset.xmid == 0.5 * (tset.xmin + tset.xmax)
        x, y = traceset2xy(tset)
        tset2 = xy2traceset(x, y, ncoeff=tset.ncoeff,
                            xjumplo=tset.xjumplo,
                            xjumphi=tset.xjumphi,
                            xjumpval=tset.xjumpval)
        assert tset2.xmin == tset.xmin
        assert tset2.xmax == tset.xmax
        assert tset2.func == tset.func
        assert np.allclose(tset2.coeff, tset.coeff)

    def test_traceset_keywords(self):
        tset = TraceSet(self.boss[1].data)
        x, y = traceset2xy(tset)
        iv = 2.0 * np.ones(x.shape, dtype=x.dtype)
        im = np.ones(x.shape, dtype=bool)
        im[1] = False
        tset2 = TraceSet(x, y, func='chebyshev', invvar=iv, maxiter=20,
                         inmask=im)
        assert tset2.func == 'chebyshev'

    def test_traceset_bad(self):
        with pytest.raises(PydlutilsException):
            _ = TraceSet(1, 2, 3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the cooling directory in idlutils.
"""
from numpy import interp
from astropy.io import ascii
from astropy.utils.data import get_pkg_data_contents


def read_ds_cooling(fname, logT=None):
    """Read in the `Sutherland & Dopita (1993)
    <https://ui.adsabs.harvard.edu/abs/1993ApJS...88..253S/abstract>`_ cooling function.

    Parameters
    ----------
    fname : { 'm-00.cie', 'm-05.cie', 'm+05.cie', 'm-10.cie', 'm-15.cie', 'm-20.cie', 'm-30.cie', 'mzero.cie' }
        Name of the data file to read.
    logT : :class:`numpy.ndarray`, optional
        If provided, values will be interpolated to the provided values.  If
        not provided, the values in the data files will be returned.

    Returns
    -------
    :class:`tuple`
        A tuple containing `logT` and `loglambda`, respectively.

    Raises
    ------
    :exc:`ValueError`
        If the input file name is invalid.

    Notes
    -----
    The data have been retrieved from
    https://www.mso.anu.edu.au/~ralph/data/cool/ and stored in the package.

    Examples
    --------
    >>> from pydl.pydlutils.cooling import read_ds_cooling
    >>> logT, loglambda = read_ds_cooling('m-15.cie')
    >>> logT[0:5] # doctest: +NORMALIZE_WHITESPACE
    array([4.  , 4.05, 4.1 , 4.15, 4.2 ])
    >>> loglambda[0:5] # doctest: +NORMALIZE_WHITESPACE
    array([-26.  , -24.66, -23.52, -22.62, -22.11])
    """
    if fname not in ('m-00.cie', 'm-05.cie', 'm+05.cie', 'm-10.cie',
                     'm-15.cie', 'm-20.cie', 'm-30.cie', 'mzero.cie'):
        raise ValueError('Invalid value for data file: {0}'.format(fname))
    coolingfile = get_pkg_data_contents('data/cooling/' + fname)
    data = ascii.read(coolingfile.split('\n')[2:], delimiter='\t')
    if logT is None:
        return (data['log(T)'].data, data['log(lambda net)'].data)
    else:
        loglambda = interp(logT, data['log(T)'].data,
                           data['log(lambda net)'].data)
        return (logT, loglambda)



try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


# Repeat this from specutils/conftest.py so tox picks it up.
def pytest_configure(config):

    if ASTROPY_HEADER:
        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES.pop('h5py', None)
        PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
        PYTEST_HEADER_MODULES['PyDL'] = 'pydl'

        from pydl import __version__
        TESTED_VERSIONS['pydl'] = __version__



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-


def median(array, width=None, axis=None, even=False):
    """Replicate the IDL ``MEDIAN()`` function.

    Parameters
    ----------
    array : array-like
        Compute the median of this array.
    width : :class:`int`, optional
        Size of the neighborhood in which to compute the median (*i.e.*,
        perform median filtering).  If omitted, the median of the whole
        array is returned.
    axis : :class:`int`, optional
        Compute the median over this axis for a multi-dimensional array.  If
        ommitted, the median over the entire array will be returned.  If
        set, this function will behave as though `even` is ``True``.
    even : :class:`bool`, optional
        If set to ``True``, the median of arrays with an even number of elements
        will be the average of the middle two values.

    Returns
    -------
    array-like
        The median of the array.

    Raises
    ------
    :exc:`ValueError`
        If `width` is set, and the input `array` is not 1 or 2 dimensional.

    Notes
    -----
    * For arrays with an even number of elements, the :func:`numpy.median`
      function behaves like ``MEDIAN(array, /EVEN)``, so the absence of
      the `even` keyword has to turn *off* that behavior.
    * For median filtering, this uses :func:`scipy.signal.medfilt` and
      :func:`scipy.signal.medfilt2d` under the hood, but patches up the
      values on the array boundaries to match the return values of the
      IDL ``MEDIAN()`` function.

    References
    ----------
    https://www.nv5geospatialsoftware.com/docs/median.html
    """
    import numpy as np
    from scipy.signal import medfilt, medfilt2d
    if width is None:
        if axis is None:
            f = array.flatten()
            if f.size % 2 == 1 or even:
                return np.median(array)
            else:
                i = f.argsort()
                return f[i[f.size//2]]
        else:
            return np.median(array, axis=axis)
    else:
        if array.ndim == 1:
            medarray = medfilt(array, min(width, array.size))
            istart = int((width - 1)/2)
            iend = array.size - int((width + 1)/2)
            i = np.arange(array.size)
            w = (i < istart) | (i > iend)
            medarray[w] = array[w]
            return medarray
        elif array.ndim == 2:
            medarray = medfilt2d(array, min(width, array.size))
            istart = int((width-1)/2)
            iend = (array.shape[0] - int((width+1)/2), array.shape[1] - int((width+1)/2))
            i = np.arange(array.shape[0])
            j = np.arange(array.shape[1])
            w = ((i < istart) | (i > iend[0]), (j < istart) | (j > iend[1]))
            medarray[w[0], :] = array[w[0], :]
            medarray[:, w[1]] = array[:, w[1]]
            return medarray
        else:
            raise ValueError("Invalid number of dimensions for input array!")
